/*
 * imbuild.c -- Build a fits 2D image from a sequence of 1D spectra
 *
 * John Kielkopf
 * kielkopf@louisville.edu
 * 
 *
 * Current Version 1.0 -- February 9, 2006
 *
 *   February 9, 2006 : Version 1.0 : First version 
 *   
 */

 /* Include */

#include <stdio.h>              /* Standard io library */
#include <math.h>               /* Math library */
#include <signal.h>             /* Error handling */
#include <stdlib.h>             /* Number conversion */
#include <string.h>             /* String functions */
#include <time.h>               /* Date and time */
#include <sys/types.h>          /* Typedefs */
#include <sys/stat.h>           /* Structure returned by stat */
#include <fcntl.h>              /* Flags for read and write */
#include "fitsio.h"             /* Fits headers */


/* Global constants */

#define NW 512                  /* Image width */
#define NH 512                  /* Image height */


/* Global data */

unsigned short array[8192][512];  /* Actual image array after scaling */


/* Function prototypes */

void write_fits_image (char *imgfile);
void printfitserror(int status);      /* Handles errors in writeimage */

int main(int argc, char *argv[])
{
    fitsfile *outfptr; 
    FILE *fpfft; 
    int k;                            /* File counter */
    int i;                            /* Character counter */
    char buf[101];                    /* FFT line buffer */ 
    double freq,signal;               /* Frequency and signal in fft */
    double fft[8192][512];            /* Buffer for fft spectral data */
    double maxsignal;                 /* Maximum signal in fft array */

    printf("Arguments found: %d\n",argc);
    if (argc < 4) { 
      printf("Usage: out_image in_data ...\n");
      printf("\n");
      printf("Build a 2D image file from 1D data\n");
      printf("\n");
      printf("Example: \n");
      printf("  imbuild spectrum.fits fft1.dat fft2.dat ...\n");
      return(0);
    }
    
    /* Initialize the image array */
    /* Here k is the row counter and i is the column counter */
    
    for (k = 0; k < 8192; k++)
    {   
      for (i = 0; i < 512; i++)
      {
        array[k][i] = 0;
        fft[k][i] = 0.;
      }
    }    

    /* Initialize maxsignal used for scaling */
    
    maxsignal = 0.001;

    /* Read each file and parse the spectra to build an image */

    for (k = 0; k < (argc - 2); k++)
    {  
    
      /* Open input fft data */
    
      printf("Opening file %s\n",argv[k+2]);
      fpfft = fopen(argv[k+2],"r");
      if (fpfft == NULL ) 
      { 
        printf("Unable to open %s \n", argv[k+2]);
      }

      /* Read data from the file and move it an image array buffer */
      
      for( i=0 ;  (fgets(buf,100,fpfft) != NULL) & (i < NW) ; i++ )
      {
        sscanf(buf,"%lf %lf",&freq, &signal);    
        fft[k][i] = signal;
      } 
      fclose(fpfft);           
    }
    
    /* Scale the data to an integer image array */

    for (k = 0; k < 8192; k++)
    {   
      for (i = 0; i < 512; i++)
      {
        array[k][i] = (unsigned short) 32767.*(fft[k][i])/maxsignal ;
      }
    }    

    write_fits_image(argv[1]);

}

/* Create fits file containing a 2-d spectral image of array[8192][512] */

void write_fits_image( char *imgfile )

{
    fitsfile *fptr;       /* Pointer to the fits file, defined in fitsio.h */
    int status;
    long  fpixel, nelements, exposure;

    /* Initialize FITS image parameters */

    int bitpix   =  USHORT_IMG;      /* 16-bit unsigned short pixel values  */
    long naxis    =   2;             /* 2-dimensional image                 */    
    long naxes[2] = { 512, 8192 };   /* image is 512 wide by 8192 rows      */
    
    /* Delete the old file if it already exists  */
    /*                                           */
    /* remove(imagename);                        */ 
    
    /* Initialize status before calling fitsio routines */ 

    status = 0;   
          
    /* Create new FITS file  */
    
    if (fits_create_file(&fptr, imgfile, &status)) printfitserror( status );           

    /* Write the required keywords for the primary array image.       */
    /* Since bitpix = USHORT_IMG, this will cause cfitsio to create   */
    /*   a FITS image with BITPIX = 16 (signed short integers) with   */
    /*   BSCALE = 1.0 and BZERO = 32768.  This is the convention that */
    /*   FITS uses to store unsigned integers.  Note that the BSCALE  */
    /*   and BZERO keywords will be automatically written by cfitsio  */
    /*   in this case.                                                */

    if ( fits_create_img(fptr,  bitpix, naxis, naxes, &status) )
         printfitserror( status );          

    fpixel = 1;                               /* First pixel to write      */
    nelements = naxes[0] * naxes[1];          /* Number of pixels to write */

    /* Write the array of unsigned integers to the FITS file */
    
    if ( fits_write_img(fptr, TUSHORT, fpixel, nelements, array[0], &status) )
        printfitserror( status );            

    /* Write optional keywords to the header here                  */
    /* Note that the ADDRESS of the value is passed in the routine */
    /*                                                             */
    /* if ( fits_update_key(fptr, TLONG, "EXPOSURE", &exposure,    */
    /*     "Seconds", &status) )                                   */
    /*     printfitserror( status );                               */

    if ( fits_close_file(fptr, &status) )                /* close the file */
         printfitserror( status );           

    return;
}



/* Print out cfitsio error messages  */

void printfitserror( int status )

{

    if (status)
    {
       fits_report_error(stderr, status); /* print error report */
    }
    return;
}



