/*
 * imrows.c -- Extract a sum of columns from a fits ccd 2D image
 *
 * John Kielkopf
 * kielkopf@louisville.edu
 * 
 *
 * Current Version 1.0 -- February 12, 2006
 *
 *   February 12, 2006 : Version 1.0 
 *   
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "fitsio.h"

int main(int argc, char *argv[])
{
    fitsfile *afptr;  
    FILE *fpdata;   
    int status = 0;  
    int atype, anaxis, check = 1, ii, jj;
    int startcol, stopcol, totpix;
    int nx, ny;
    long npixels = 1, firstpix[2] = {1,1};
    long anaxes[2] = {1,1};
    double *apix, *sum;

    if (argc != 5) { 
      printf("Usage: imcols in_image out_data start_col stop_col\n");
      printf("\n");
      printf("Write a data file with a sum of fits image columns\n");
      printf("\n");
      printf("Example: \n");
      printf("  imcols image.fits cols.dat 100 200\n");
      return(0);
    }
    
    startcol = atoi(argv[3]);
    stopcol = atoi(argv[4]);

    /* Open input image */
    
    fits_open_file(&afptr, argv[1], READONLY, &status); 

    /* Read dimensions */

    fits_get_img_dim(afptr, &anaxis, &status);  
    fits_get_img_size(afptr, 2, anaxes, &status);

    nx = anaxes[0];
    ny = anaxes[1];

    /* Report errors if any */

    if (status) {
       fits_report_error(stderr, status); 
       return(status);
    }

    /* Check the image sizes */
    
    if (anaxis != 2) 
    {
       printf("Error: only a 2-D image is supported\n");
       fits_close_file(afptr, &status);
       return(1);    
    }
    
    /* Check the column range */
    
    if (stopcol < startcol) 
    {
       printf("Initial column must be less than or equal to last column\n");
       fits_close_file(afptr, &status);
       return(1);    
    }
    
    if (startcol < 1) 
    {
       printf("Initial column must be greater than 1\n");
       fits_close_file(afptr, &status);
       return(1);    
    }

    if (stopcol > nx) 
    {
       printf("Last column must be less than %d\n for this image", anaxes[0]);
       fits_close_file(afptr, &status);
       return(1);    
    }


    /* Open a file to write the sum data */

    fpdata = fopen(argv[2],"w");  
  
    /* Number of pixels to read in each row */
    
    if (fpdata == NULL ) 
    { 
      printf("Unable to open output file for data\n");
      fits_close_file(afptr, &status);
      return(1);
    }
    
    
    /* Allocate memory for one row */ 

    apix = (double *) malloc(nx * sizeof(double)); 

    if (apix == NULL ) 
    {
      printf("Image row memory allocation error\n");
      return(1);
    }


    /* Allocate memory for the sum */ 

    sum = (double *) malloc(ny * sizeof(double)); 

    if (sum == NULL ) 
    {
      printf("Sum data memory allocation error\n");
      return(1);
    }
    
    /* Initialize the sum array */
    
    for (jj = 0; jj < ny; jj++) 
    {
      sum[jj] = 0.;                      
    }         

    /* Check size of image */
    
    totpix = nx * ny;
    if (totpix == 0)
    {
      printf("The  image is empty\n");
      return(1);
    }
    
    /* Start with first pixel in each row */
    
    firstpix[0] = 1;  

    /* Process image one row at a time using row index  jj */
    
    jj = 0;
    npixels = nx;

    for (firstpix[1] = 1; firstpix[1] <= ny; firstpix[1]++)
    {  

      /* Give starting pixel coordinate and number of pixels to read */
       
      if (fits_read_pix(afptr, TDOUBLE, firstpix, npixels,
           NULL, apix, NULL, &status))
        break;   

      for (ii = startcol - 1; ii < stopcol; ii++) 
      {
        sum[jj] += apix[ii];                      
      }
      jj++;
    }
    
    /* Write the data file */
    
    for (jj = 0; jj < ny; jj++) 
    {      
      fprintf(fpdata, "%d  %lf\n", jj + 1, sum[jj]);   
    }
    
    
    free(apix);
    free(sum);

    fits_close_file(afptr, &status);
    fclose(fpdata);

    /* Report any error message */
 
    if (status) fits_report_error(stderr, status); 
    return(status);
}

