/* 
 * -----------------------------------------------------------------------
 * -     Astronomical CCD Camera Control with an Motif User Interface    -
 * -----------------------------------------------------------------------
 *
 * Copyright 2002,2004,2005,2006 John Kielkopf
 * kielkopf@louisville.edu
 * 
 * Source code, but not the camera library, is distributed under the terms 
 * of the General Public License (see LICENSE)
 *
 * This program is offered without warranty of any kind. 
 *
 * Date: September 28, 2007
 * Version: 2.0.9
 *
 * History:
 * 
 *   March 9, 2002
 *     Version 1.0.0
 *     Support for SBIG cameras using the first public SBIG library
 *     Basic panel for operation with ds9
 *
 *   May 22, 2002
 *     Version 1.1.0
 *     Updated to use libudrv 4.35
 *
 *   May 24, 2004
 *     Version 1.2.0 
 *     USB support
 *
 *   March 1, 2005
 *     Version 1.4.0
 *     Full FITS file support
 *
 *   July 3, 2005
 *     Version 1.5.0
 *     Support for all filterwheels
 *
 *   January 1, 2006
 *     Version 2.0.0
 *     AO control
 *
 *   October 10, 2006
 *     Version 2.0.7
 *     Single, Focus and Multi imaging modes
 *
 *   October 14, 2006
 *     Version 2.0.8
 *     Guide mode with region from ds9
 *
 *   April 7, 2007
 *     Version 2.0.8b
 *     Incremented month in gmt stored by FITS so that it is 1-12
 *
 *   September 28, 2007
 *     Version 2.1
 *     Incremented version number in preparation for release of this version
 *     GUI modified with addition of an Edit menu and changes to filter selection
 *     Time delay integration (TDI) imaging mode
 *     Guide mode implementation completed and added to menu system
 *     Error checking added on startup for presence of the CCD
 *
 */

#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include "fitsio.h"

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h> 
#include <Xm/CascadeB.h>
#include <Xm/MessageB.h>
#include <Xm/Separator.h>
#include <Xm/SelectioB.h>

#include "sbigudrv.h"


/* Numbers */

#define PI  3.14159265359


/* Menubar identifiers */

#define OK             1
#define CANCEL         2
#define EXIT           3
#define BASENAME       4
#define SETIMAGE       5
#define SETTRACK       6
#define SETTDI         7
#define SETTEMPERATURE 8
#define SETSINGLEMODE      9
#define SETFOCUSMODE      10
#define SETMULTIMODE      11
#define SETTDIMODE        12
#define SETSELFGUIDEMODE  13


/* Cooling dialog flags */

#define COOL_ON   1
#define COOL_OFF  0


/* Map sky to +x=1, -x=2, +y=4, -y=8 relays */

#define EAST      1 
#define WEST      2
#define NORTH     4
#define SOUTH     8 


/* Feedback modes */

#define TRACKING  0
#define ADAPTIVE  1


/* Imaging modes */

#define SINGLE    0
#define FOCUS     1
#define MULTI     2
#define TDI       3
#define GUIDE     4 


/* CCD selection */

#define IMAGE            0
#define TRACK            1


XtAppContext context;
XmStringCharSet char_set=XmSTRING_DEFAULT_CHARSET;

Widget menu_bar;
Widget file_menu;
Widget exit_item;
Widget edit_menu;
Widget basename_item;
Widget set_image_exp_item;
Widget set_track_exp_item;
Widget set_temperature_item;
Widget set_tdi_item;

Widget mode_menu;
Widget single_item;
Widget focus_item;
Widget multi_item;
Widget tdi_item;
Widget selfguide_item;

Widget toplevel;
Widget image_exposure_dialog;
Widget track_exposure_dialog;
Widget temperature_dialog;
Widget basename_dialog;
Widget tdi_dialog;


Widget button[3][3];
Widget command[5];
Widget toggle[4];
Widget label_info[6];
Widget show_info[6];
Widget form;
Widget sep_btns;
Widget sep_status;
Widget radio_box;
XtIntervalId ccd_timeout_id;
XtPointer ccd_timeout_data_ptr;


Widget make_menu();              /* creates a menu on the menu bar */
Widget make_menu_item();         /* adds an item into the menu */
Widget make_menu_item_toggle();  /* adds a toggle button item into the menu */


/* labels for the buttons top to bottom then left to right */

static char btn_name[][6] = {
  "NoOp", 
  "East",
  "NoOp",
  "North",
  "NoOp",
  "South",
  "NoOp",
  "West",
  "NoOp"
};

static char cmd_name[][30] = {
  "Expose Image CCD",
  "Expose Image CCD Dark Frame",
  "Expose Track CCD ", 
  "Interrupt Current Exposure",
  "Change Filter"
};

static char trk_name[][20] = {
  "Tracking",
  "Adaptive"
 };

static char infotext[][24] = {
  "Track exposure",
  "Image exposure",
  "CCD temperature",
  "TDI parameters",
  "Filter",
  "Status"
};

static char infodata[][24] = {
  "-",
  "-",
  "-",
  "-",
  "-",
  "Camera not ready"
};


/* Filterwheel support */

#define NO_FILTERWHEEL 0xFFFF

/* Type of filterwheel: set CFWSEL_CFW8, _CFW10, or _CFWL */
/* If no filterwheel is present: set NO_FILTERWHEEL */
/* CFWSEL_AUTO will not work with the CFW10 */

static CFW_MODEL_SELECT    cfw_model_select = CFWSEL_CFWL ;

/* Preferred startup filter position */
/* When camera is initialized software will call for this position */
/* In the STL series cameras with SBIG's system, filter 4 will be clear */
/* In the UBVRI system filter 3 will be V */

int filter=3;  

/*  Maxmimum number of filters to allow for this filterwheel */

int filtermax=5;  

/* For the CFW10 this should be set to 10 */

/* Filter names */
/* Maximum 24 characters */
/* FITS FILTER header will include this label */

/* The following are examples for specific filter arrangements */
/* Uncomment one and edit to your needs */

/*

static char cfw_filter_name[][24] = {
  " (1)              ",
  " (2)              ",
  " (3)              ",
  " (4)              ",
  " (5)              ",
  " (6)              ",  
  " (7)              ",  
  " (8)              ",
  " (9)              ", 
  "(10)              "
 };

static char cfw_filter_name[][24] = {
  " (1) C        ",
  " (2) L        ",
  " (3) B (RGB)  ",
  " (4) G (RGB)  ",
  " (5) R (RGB)  ",
  " (6) U (UBVRI)",  
  " (7) B (UBVRI)",  
  " (8) V (UBVRI)",
  " (9) R (UBVRI)", 
  "(10) I (UBVRI)"
 };




static char cfw_filter_name[][24] = {
  " (1) R (580-680)  ",
  " (2) G (475-580)  ",
  " (3) B (390-490)  ",
  " (4) C (350-1100) ",
  " (5) L (390-680)  ",
  " (6)              ",  
  " (7)              ",  
  " (8)              ",
  " (9)              ", 
  "(10)              "
 };
 

*/

static char cfw_filter_name[][24] = {
  " (1) I (725-1025) ",
  " (2) R (570-690)  ",
  " (3) V (490-580)  ",
  " (4) B (375-490)  ",
  " (5) U (330-390)  ",
  " (6)              ",  
  " (7)              ",  
  " (8)              ",
  " (9)              ", 
  "(10)              "
 };
 


/*

static char cfw_filter_name[][24] = {
  " (1) I (700-1100) ",
  " (2) Ha (654-658) ",
  " (3) R (UBVRI)    ",
  " (4) V (UBVRI)    ",
  " (5) C (350-1100) ",
  " (6)              ",  
  " (7)              ",  
  " (8)              ",
  " (9)              ", 
  "(10)              "
 };
 



*/ 

static int startup_status_flag=1;  /* Initialize GUI with defaults */
 
void create_menus();       /* Creates all menus for this program    */

void menuCB();             /* Callback routine used for all menus   */

void track_exposure_dialogCB();  /* Callback functions for dialog boxes  */
void image_exposure_dialogCB();  
void basename_dialogCB();        
void temperature_dialogCB();     
void tdi_dialogCB();           

void buttonCB();           /* Callback function for the 9 paddle      */
                           /* buttons called when one of the paddle   */
                           /* buttons is clicked or unclicked.        */
                           
void feedbackCB();         /* Callback function for the feedback toggles */
                           
void commandCB();          /* Callback function for the 8 command      */
                           /* buttons called when one of the command   */
                           /* buttons is clicked.                      */

void ccd_timeout_handler();       /* Handles status data flow and tracking */
void ccd_status();                /* Captures status data from ccd */
void ccd_cool_on();               /* Cooling on at ccd_setpoint_temperature */
void ccd_cool_off();              /* Leave setpoint unchanged but turn off */
void ccd_relay_on();              /* Turns on trackrelay for 10 s */
void ccd_relay_off();             /* Turns off trackrelay */

void ao_tip_tilt(int ix, int iy);  /* Set tip and tilt of ao device */
void ao_test(int aomag, int aodelay, int aocycles);  /* Test ao device */

void increment_filter();          /* Increment the filter number and check bounds */
void set_filter();                /* Set filter for CFW */
void read_ccd();                  /* Test for completion and read as needed */
void start_ccd_exposure();        /* Begin exposure for chosen ccd */

void input_track_time();          /* Input track exposure time */
void input_image_time();          /* Input image exposure time */
void input_ccd_setpoint();        /* Input and set ccd setpoint temperature */
void increment_filter();          /* Increment filter number and select filter */

void update_track_exposure();     /* Update track exposure time display */
void update_image_exposure();     /* Update image exposure time display */
void update_ccd_temperature();    /* Update ccd thermistor display */
void update_tdi_parameters();     /* Update tdi info display */
void update_filter();             /* Update filter number display */
void update_status();             /* Update status buffer string display */
void update_all();                /* Update all display buffers */
void start_ds9();                 /* Initiate ds9 display */
void update_ds9_image();          /* xpa to ds9: reread image.fits */
void update_ds9_track();          /* xpa to ds9: reread track.fits */
void get_ds9_region();            /* xpa from ds9: send region */

/* Tracking parameters */

int feedbackmode=TRACKING;        /* camera control of image position */
int trackrelay;                   /* relay choice */
int ao_x=2048;                    /* adaptive optics tip  defaults to center */
int ao_y=2048;                    /* adaptive optics tilt defaults to center */
int subarea_x=0;                  /* x at corner of subarea readout */
int subarea_y=0;                  /* y at corner of subarea readout */
int subarea_width=8192;           /* width of subarea initially very large */
int subarea_height=8192;          /* height of subarea initially very large */
double region_x;                  /* x center of selected region */
double region_y;                  /* y center of selected region */
double region_dx;                 /* height of selected region */
double region_dy;                 /* width of selected region */

/* Files */

char basename[64];                /* Basename for multi-image files */
int imagenumber=0;                /* Serial number for multi-image files */
char imagename[64];               /* File name for a multi-image file */

/* Operating parameters */

int exposure_in_progress=FALSE;   /* exposure flag */
int image_mode=SINGLE;            /* image acquisition mode */

double track_exposure_time=0.2;   /* default tracking ccd exposure time in seconds */
double image_exposure_time=1.;    /* default imaging ccd exposure time in seconds */
double ccd_temperature;           /* CCD temperature in C */
double ccd_setpoint_temperature=2.;   /* CCD setpoint temperature in C */
double ccd_setpoint_reading;      /* CCD setpoint read from camera */
double ccd_cooler_power;          /* power level as percent of full power */

/* Initial TDI parameters will be read from ini file */

double declination = 0.;          /* center-of-field declination in degrees */
double ccd_scale = 0.5372967;     /* ccd scale in seconds of arc per pixel */
int ccd_line_time = 12332;        /* ccd line read time in microseconds */
int tdi_delay = 1;                /* tdi delay time in microseconds */
int tdi_height = 6000;            /* tdi single frame image height */

/* Buffers */

char buf[256];                   /* general purpose character buffer */
char status_buf[24];             /* short buffer for status string */

/* Flags and interface values */

static int interrupt=FALSE;      /* interrupt flag TRUE or FALSE */
static int restart=FALSE;        /* multiple exposure flag */

unsigned long ccd_timeout_interval=500;   /* timeout service interval*/

static short ccd_temp[4096];      /* CCD thermistor lookup table */


static CAMERA_TYPE  camera_type = NO_CAMERA;
static GetCCDInfoResults0  info0;
static GetCCDInfoResults0  info1;
static GetCCDInfoResults2  info2;
static GetCCDInfoResults3  info3;
static int image_in_progress, track_in_progress;
void write_fits(char *filename, int w, int h, unsigned short *data);
void show_cfitsio_error(int status);
void show_error(char *string, int error);
void establish_link(int camera_device);
void get_ccd_info(void);

static void usage(FILE *stream);     /* List command line arguments */ 
int argmatch(char *s1, char *s2, int atleast);  /* Compare arguments */

static int camera_device = DEV_USB1;            /* default camera device */
static int camera_ipaddress = 0xc0a80001;       /* default hex ip address */
int result;
GetDriverInfoParams  in;
GetDriverInfoResults0 out;

int ccd=IMAGE;               /*  ccd mode */
int readout_ccd=IMAGE;       /*  active ccd during readout */
int is_dark=FALSE;           /*  dark frame flag  */


/* GUI main starts here */

int main(argc,argv)
    int argc; 
    char *argv[];
{
    Arg al[10];
    int ac,i;
    int x,y,btn_number,cmd_number;
    int info_number;
    EventMask mask;

/* Check the command line and parse arguments */

    if (argc >= 1) {
	for (i = 1; i < argc; i++) {
	    if (argv[i][0] == '-' && argv[i][1] != '\0') {
		if (argmatch(argv[i], "-usb", 4)) {
                    camera_device = DEV_USB ;
		}
                else if (argmatch(argv[i], "-usb1", 5)) {
                    camera_device = DEV_USB1;
		}
                else if (argmatch(argv[i], "-usb2", 5)) {
                    camera_device = DEV_USB2;
		} 
                else if (argmatch(argv[i], "-usb3", 5)) {
                    camera_device = DEV_USB3;
                }    
                else if (argmatch(argv[i], "-usb4", 5)) {
                    camera_device = DEV_USB4;
                }                
                else if (argmatch(argv[i], "-lpt1", 5)) {
                    camera_device = DEV_LPT1;
		}
                else if (argmatch(argv[i], "-lpt2", 5)) {
                    camera_device = DEV_LPT2;
		} 
                else if (argmatch(argv[i], "-lpt3", 5)) {
                    camera_device = DEV_LPT3;
                }
                else if (argmatch(argv[i], "-eth", 4)) {
                    camera_device = DEV_ETH;
                }
                
                else {
                    usage(stderr);
                }
            }
            
        }
    }                            
    else {
        usage(stderr);
    }

/* Initialize the CCD interface and make contact */

    SBIGUnivDrvCommand(CC_OPEN_DRIVER, NULL, NULL);
    in.request = 0;
    result = SBIGUnivDrvCommand(CC_GET_DRIVER_INFO, &in, &out);
    establish_link(camera_device);
    get_ccd_info();

/* Start ds9 for image display */
/* Prototypes ds9.fits, image.fits and track.fits will be in cwd */
    
    start_ds9();

/* Create the user interface */

    /* create the toplevel shell */

    toplevel = XtAppInitialize(&context,"",NULL,0,&argc,argv,
        NULL,NULL,0);


    /* set the default size of the window */ 

    ac=0;
    XtSetArg(al[ac],XmNwidth,400); ac++;
    XtSetArg(al[ac],XmNheight,600); ac++;
    XtSetValues(toplevel,al,ac);


    /* create a form widget */

    ac=0;
    form=XmCreateForm(toplevel,"form",al,ac);
    XtManageChild(form);


    /* create the menu bar and attach it to the form */
    
    ac=0;
    XtSetArg(al[ac],XmNtopAttachment,XmATTACH_FORM); ac++;
    XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++;
    XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;
    menu_bar=XmCreateMenuBar(form,"menu_bar",al,ac);
    XtManageChild(menu_bar);


    /* note that locations of buttons are in percent of window dimension */

    /* create a radio box to hold the feedback toggles and attach it to form */
    
    ac=0;
    XtSetArg(al[ac],XmNorientation,XmHORIZONTAL); ac++;
    XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNtopPosition,5); ac++; 
    XtSetArg(al[ac],XmNleftAttachment,XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNleftPosition,30); ac++; 
    XtSetArg(al[ac],XmNrightAttachment,XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNrightPosition,70); ac++; 
    radio_box=XmCreateRadioBox(form,"radio_box",al,ac);
    XtManageChild(radio_box);
    

    /* create feedback mode toggles */
    
    for (x=0; x<2; x++)
    {
         ac=0;
         XtSetArg(al[ac],XmNlabelString,
           XmStringCreate(trk_name[x],char_set)); ac++;
         toggle[x]=XmCreateToggleButton(radio_box,"toggle",al,ac);
         XtManageChild(toggle[x]);
         XtAddCallback (toggle[x], 
           XmNvalueChangedCallback, feedbackCB, (int *) x);
    }     

    /* Set the tracking togglebutton to initial value of feedbackmode */

    XmToggleButtonSetState (toggle[0], True, False);
 
 

    /* set up the handpaddle buttons, add event handlers, attach to form. */

    for (x=0; x<3; x++)
    {
         for (y=0; y<3; y++)
         {
              btn_number=3*x+y;
              if ((btn_number==1) | (btn_number==3) | (btn_number==5) | (btn_number==7) ) 
              {
              ac=0;
              XtSetArg(al[ac],XmNlabelString,
                  XmStringCreate(btn_name[btn_number],char_set)); ac++;
              XtSetArg(al[ac],XmNleftAttachment,
                  XmATTACH_POSITION); ac++;
              XtSetArg(al[ac],XmNleftPosition,27+x*16); ac++;
              XtSetArg(al[ac],XmNrightAttachment,
                  XmATTACH_POSITION); ac++;
              XtSetArg(al[ac],XmNrightPosition,41+x*16); ac++;
              XtSetArg(al[ac],XmNtopAttachment,
                  XmATTACH_POSITION); ac++;
              XtSetArg(al[ac],XmNtopPosition,12+y*9); ac++;
              XtSetArg(al[ac],XmNbottomAttachment,
                  XmATTACH_POSITION); ac++;
              XtSetArg(al[ac],XmNbottomPosition, 20+y*9); ac++;
              button[x][y]=XmCreatePushButton(form,"label",al,ac);
              mask = ButtonReleaseMask | ButtonPressMask;
              XtAddEventHandler(button[x][y],mask,False,
                  buttonCB,(int *) btn_number);              
              XtManageChild(button[x][y]);


             }
         }
    }
 

    /* Create a separator widget and attach it to the form */

    ac=0;
    XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNtopPosition,42); ac++;
    XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
    XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
    XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;  
    sep_btns=XmCreateSeparator(form,"sep",al,ac);
    XtManageChild(sep_btns);   
 


    /* Set up the information label widgets and attach them to the form */

    for  (x=0; x<2; x++)
    {
      for (y=0; y<3; y++)
      {
        info_number=2*y+x;
        ac=0;
        XtSetArg(al[ac],XmNlabelString,
            XmStringCreate(infotext[info_number],char_set)); ac++;
        XtSetArg(al[ac],XmNleftAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNleftPosition,5+x*50); ac++;
        XtSetArg(al[ac],XmNrightAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNrightPosition,45+x*50); ac++;
        XtSetArg(al[ac],XmNtopAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNtopPosition,45+y*6); ac++;
        XtSetArg(al[ac],XmNbottomAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNbottomPosition,47+y*6); ac++;
        label_info[info_number]=XmCreateLabel(form,"label",al,ac);
        XtManageChild(label_info[info_number]); 
      }
    }


    /* Set up the information data widgets and attach them to the form */

    for  (x=0; x<2; x++)
    {
      for (y=0; y<3; y++)
      {
        info_number=2*y+x;
        ac=0;
        XtSetArg(al[ac],XmNlabelString,
            XmStringCreate(infodata[info_number],char_set)); ac++;
        XtSetArg(al[ac],XmNleftAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNleftPosition,5+x*50); ac++;
        XtSetArg(al[ac],XmNrightAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNrightPosition,45+x*50); ac++;
        XtSetArg(al[ac],XmNtopAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNtopPosition,47+y*6); ac++;
        XtSetArg(al[ac],XmNbottomAttachment,
            XmATTACH_POSITION); ac++;
        XtSetArg(al[ac],XmNbottomPosition,49+y*6); ac++;
        show_info[info_number]=XmCreateLabel(form,"label",al,ac);
        XtManageChild(show_info[info_number]); 
      }
    }


    /* Create a separator widget and attach it to the form */

    ac=0;
    XtSetArg(al[ac],XmNtopAttachment,XmATTACH_POSITION); ac++;
    XtSetArg(al[ac],XmNtopPosition,64); ac++;
    XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++; 
    XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;  
    XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_NONE); ac++;      
    sep_status=XmCreateSeparator(form,"sep",al,ac);
    XtManageChild(sep_status);   
    

    /* The following provide the push button commands */

    /* set up the command buttons and attach them to the form */

    for (y=0; y<5; y++)
    {
              cmd_number=y;
              ac=0;
              XtSetArg(al[ac],XmNlabelString,
                  XmStringCreate(cmd_name[cmd_number],char_set)); ac++;
              XtSetArg(al[ac],XmNleftAttachment,
                  XmATTACH_POSITION); ac++;
              XtSetArg(al[ac],XmNleftPosition,20); ac++;
              XtSetArg(al[ac],XmNrightAttachment,
                  XmATTACH_POSITION); ac++;
              XtSetArg(al[ac],XmNrightPosition,80); ac++;
              XtSetArg(al[ac],XmNtopAttachment,
                  XmATTACH_POSITION); ac++;
              XtSetArg(al[ac],XmNtopPosition,68+y*6); ac++;
              XtSetArg(al[ac],XmNbottomAttachment,
                  XmATTACH_POSITION); ac++;
              XtSetArg(al[ac],XmNbottomPosition, 73+y*6); ac++;
              command[y]=XmCreatePushButton(form,"label",al,ac);
              XtManageChild(command[y]);
              XtAddCallback(command[y],XmNactivateCallback,
                  commandCB,(int *) cmd_number);
     }

    /* create a prompt dialog for image CCD exposure input but leave it unmanaged */

    ac=0;
    XtSetArg(al[ac], XmNselectionLabelString,
        XmStringCreateLtoR("Enter time (s): ",char_set)); ac++;
    XtSetArg(al[ac], XmNdialogTitle,
        XmStringCreateLtoR("Imaging CCD Exposure",char_set)); ac++;
    image_exposure_dialog=XmCreatePromptDialog(toplevel,"dialog",al,ac);
    XtAddCallback(image_exposure_dialog, XmNokCallback, image_exposure_dialogCB, (int *) OK);
    XtAddCallback(image_exposure_dialog, XmNcancelCallback, image_exposure_dialogCB, (int *) CANCEL);
    XtUnmanageChild (XmSelectionBoxGetChild (image_exposure_dialog, XmDIALOG_HELP_BUTTON));
   
    /* create a prompt dialog for track CCD exposure input but leave it unmanaged */

    ac=0;
    XtSetArg(al[ac], XmNselectionLabelString,
        XmStringCreateLtoR("Enter time (s): ",char_set)); ac++;
    XtSetArg(al[ac], XmNdialogTitle,
        XmStringCreateLtoR("Tracking CCD Exposure",char_set)); ac++;
    track_exposure_dialog=XmCreatePromptDialog(toplevel,"dialog",al,ac);
    XtAddCallback(track_exposure_dialog, XmNokCallback, track_exposure_dialogCB, (int *) OK);
    XtAddCallback(track_exposure_dialog, XmNcancelCallback, track_exposure_dialogCB, (int *) CANCEL);
    XtUnmanageChild (XmSelectionBoxGetChild (track_exposure_dialog, XmDIALOG_HELP_BUTTON));

    /* create a prompt dialog for CCD temperature input but leave it unmanaged */

    ac=0;
    XtSetArg(al[ac], XmNselectionLabelString,
        XmStringCreateLtoR("Enter temperature (C): ",char_set)); ac++;
    XtSetArg(al[ac], XmNdialogTitle,
        XmStringCreateLtoR("CCD Temperature Setpoint",char_set)); ac++;
    XtSetArg(al[ac], XmNokLabelString,
        XmStringCreateLtoR("On",char_set)); ac++;
    XtSetArg(al[ac], XmNhelpLabelString,
        XmStringCreateLtoR("Off",char_set)); ac++;
    temperature_dialog=XmCreatePromptDialog(toplevel,"dialog",al,ac);
    XtAddCallback(temperature_dialog, XmNokCallback, temperature_dialogCB, (int *) COOL_ON);
    XtAddCallback(temperature_dialog, XmNcancelCallback, temperature_dialogCB, (int *) CANCEL);
    XtAddCallback(temperature_dialog, XmNhelpCallback, temperature_dialogCB, (int *) COOL_OFF);

    /* create a prompt dialog for TDI parameters but leave it unmanaged */

    ac=0;
    XtSetArg(al[ac], XmNselectionLabelString,
        XmStringCreateLtoR("Enter declination: ",char_set)); ac++;
    XtSetArg(al[ac], XmNdialogTitle,
        XmStringCreateLtoR("TDI Parameters",char_set)); ac++;
    tdi_dialog=XmCreatePromptDialog(toplevel,"dialog",al,ac);
    XtAddCallback(tdi_dialog, XmNokCallback, tdi_dialogCB, (int *) OK);
    XtAddCallback(tdi_dialog, XmNcancelCallback, tdi_dialogCB, (int *) CANCEL);
    XtUnmanageChild (XmSelectionBoxGetChild (tdi_dialog, XmDIALOG_HELP_BUTTON));
   
    /* create a prompt dialog for file basename input but leave it unmanaged */

    ac=0;
    XtSetArg(al[ac], XmNselectionLabelString,
        XmStringCreateLtoR("Enter file basename: ",char_set)); ac++;
    XtSetArg(al[ac], XmNdialogTitle,
        XmStringCreateLtoR("Filename selection",char_set)); ac++;
    basename_dialog=XmCreatePromptDialog(toplevel,"dialog",al,ac);
    XtAddCallback(basename_dialog, XmNokCallback, basename_dialogCB, (int *) OK);
    XtAddCallback(basename_dialog, XmNcancelCallback, basename_dialogCB, (int *) CANCEL);
    XtUnmanageChild (XmSelectionBoxGetChild (basename_dialog, XmDIALOG_HELP_BUTTON));
  
    /* create the menubar */

    create_menus(menu_bar);  
      
    /* add timeout to service guiding and status display */

    ccd_timeout_id = XtAppAddTimeOut(context,
      ccd_timeout_interval,ccd_timeout_handler,ccd_timeout_data_ptr);
    
    /* start the user interface */
    
    XtRealizeWidget(toplevel);
    XtAppMainLoop(context);
}

Widget make_menu_item(item_name,client_data,menu)
    char *item_name;
    int client_data;
    Widget menu;
/* adds an item into a menu in the menubar */     
{
    int ac;
    Arg al[10];
    Widget item;

    ac = 0;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR(item_name,char_set)); ac++;      
    item=XmCreatePushButton(menu,item_name,al,ac);
    XtManageChild(item);
    XtAddCallback (item,XmNactivateCallback,menuCB,(int *) client_data);      
    XtSetSensitive(item,True);
    return(item);
}


Widget make_menu_item_toggle(item_name,client_data,menu)
    char *item_name;
    int client_data;
    Widget menu;
/* Add a toggle item into a menu in the menubar */     
{
    int ac;
    Arg al[10];
    Widget item;

    ac = 0;
    XtSetArg(al[ac], XmNlabelString,
        XmStringCreateLtoR(item_name,char_set)); ac++; 
    item=XmCreateToggleButton(menu,item_name,al,ac);
    XtManageChild(item);
    XtAddCallback (item,XmNvalueChangedCallback,menuCB,(int *) client_data);      
    XtSetSensitive(item,True);
    return(item);
}

/* Creates a menu on the menu bar */

Widget make_menu(menu_name,menu_bar)
    char *menu_name; 
    Widget menu_bar;
{
    int ac;
    Arg al[10];
    Widget menu, cascade;

    menu=XmCreatePulldownMenu(menu_bar,menu_name,NULL,0);
    ac=0;
    XtSetArg(al[ac],XmNsubMenuId,menu);  ac++;
    XtSetArg(al[ac],XmNlabelString,
        XmStringCreateLtoR(menu_name,char_set)); ac++;
    cascade=XmCreateCascadeButton(menu_bar,menu_name,al,ac);      
    XtManageChild (cascade); 
    return(menu);
}


/* Create all the menubar entries and submenus */

void create_menus(menu_bar)
    Widget menu_bar;
{
    int ac;
    Arg al[10];
    
    /* create the file menu */
    file_menu = make_menu("File",menu_bar);
    exit_item = make_menu_item("Exit",EXIT,file_menu);
    
    /* create the edit menu */
    edit_menu = make_menu("Edit",menu_bar);
    set_image_exp_item = make_menu_item("Image exposure",SETIMAGE,edit_menu);
    set_track_exp_item = make_menu_item("Track exposure",SETTRACK,edit_menu);
    set_tdi_item = make_menu_item("TDI setup",SETTDI,edit_menu);
    set_temperature_item = make_menu_item("CCD temperature",SETTEMPERATURE,edit_menu);
    basename_item = make_menu_item("Image names",BASENAME,edit_menu);
    
    /* Create the mode menu */
    mode_menu = make_menu("Mode",menu_bar);
    single_item = make_menu_item_toggle("Single",SETSINGLEMODE,mode_menu);
    focus_item = make_menu_item_toggle("Focus",SETFOCUSMODE,mode_menu);
    multi_item  = make_menu_item_toggle("Multi",SETMULTIMODE,mode_menu);
    tdi_item = make_menu_item_toggle("TDI",SETTDIMODE,mode_menu);
    selfguide_item  = make_menu_item_toggle("Selfguide",SETSELFGUIDEMODE,mode_menu);
    XtSetSensitive(selfguide_item,False);
    ac=0;
    XtSetArg(al[ac],XmNradioBehavior,True); ac++;
    XtSetValues(mode_menu,al,ac);
    
    /* Set the initial image mode button state to match imagemode */
    XmToggleButtonSetState(single_item, True, False);
}


/* Callback routine used for menubar */

void menuCB(w,client_data,call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
{

    if (client_data==EXIT)     /* if exit detected, then exit */
    {
      exit(0);
    }
    
    if (client_data==SETIMAGE) /* if basename request detected, then prompt for name */
    {
      XtManageChild(image_exposure_dialog);
    }
    
    
    if (client_data==SETTRACK) /* if basename request detected, then prompt for name */
    {
      XtManageChild(track_exposure_dialog);
    }
   

    if (client_data==SETTDI) /* if TDI setup request detected, then prompt for information */
    {
      XtManageChild(tdi_dialog);
    }
 

    if (client_data==SETTEMPERATURE) /* if basename request detected, then prompt for name */
    {
      XtManageChild(temperature_dialog);
    }
   
    
    if (client_data==BASENAME) /* if basename request detected, then prompt for name */
    {
      XtManageChild(basename_dialog);
    }
    
   
    
    if (client_data==CANCEL) ;  /* noop */

    if (client_data==SETSINGLEMODE)
    {
      if( XmToggleButtonGetState(single_item) )
      {
        image_mode=SINGLE;
      }
    }
    if (client_data==SETFOCUSMODE)
    {
      if( XmToggleButtonGetState(focus_item) )
      {
        image_mode=FOCUS;
      }
    }    
    if (client_data==SETMULTIMODE)
    {
      if( XmToggleButtonGetState(multi_item) )
      {
        image_mode=MULTI;
      }
    }
    if (client_data==SETTDIMODE)
    {
      if( XmToggleButtonGetState(tdi_item) )
      {
        image_mode=TDI;
      }
    }
     
    if (client_data==SETSELFGUIDEMODE)
    {
      if( XmToggleButtonGetState(selfguide_item) )
      {
        image_mode=GUIDE;
      }
    }     
}


/* Callback function for an OK click on user input of track exposure time */

void track_exposure_dialogCB(w,client_data,call_data)
    Widget w;
    int client_data;
    XmSelectionBoxCallbackStruct  *call_data;
{    
    char *s;

    switch (client_data)
    {
       case OK:
         XmStringGetLtoR(call_data->value,char_set,&s);
         track_exposure_time=atof(s);
         update_track_exposure();
         XtFree(s);
         break;
       case CANCEL:
         break;
    }
    XtUnmanageChild(w);

}

/* Callback function for an OK click on user input of image exposure time */

void image_exposure_dialogCB(w,client_data,call_data)
    Widget w;
    int client_data;
    XmSelectionBoxCallbackStruct  *call_data;
{    
    char *s;

    switch (client_data)
    {
       case OK:
         XmStringGetLtoR(call_data->value,char_set,&s);
         image_exposure_time=atof(s);
         update_image_exposure();
         XtFree(s);
         break;
       case CANCEL:
         break;
    }
    XtUnmanageChild(w);

}

/* Callback function for user input of ccd temperature setpoint */

void temperature_dialogCB(w,client_data,call_data)
    Widget w;
    int client_data;
    XmSelectionBoxCallbackStruct  *call_data;
{    
    char *s;

    switch (client_data)
    {
       case COOL_ON:
         XmStringGetLtoR(call_data->value,char_set,&s);
         ccd_setpoint_temperature=atof(s);
         XtFree(s);
         update_ccd_temperature();
         ccd_cool_on();
         break;
       case COOL_OFF:
         ccd_cool_off();
         break;
       case CANCEL:
         break;
    }
    XtUnmanageChild(w);
}

/* Dialog for user input of file basename */

void basename_dialogCB(w,client_data,call_data)
    Widget w;
    int client_data;
    XmSelectionBoxCallbackStruct  *call_data;
{    
    char *s;
    
    switch (client_data)
    {
       case OK:
         XmStringGetLtoR(call_data->value,char_set,&s);
         strcpy(basename,s);
         XtFree(s);
         break;
       case CANCEL:
         break;
    }
    XtUnmanageChild(w);
}

/* Dialog for user input of TDI parameters */

void tdi_dialogCB(w,client_data,call_data)
    Widget w;
    int client_data;
    XmSelectionBoxCallbackStruct  *call_data;
{    
    char *s;

    switch (client_data)
    {
       case OK:
         XmStringGetLtoR(call_data->value,char_set,&s);
         declination=atof(s);
         if ( ((int) declination) > 85. ) break; 
         tdi_delay = (int) 66667.*ccd_scale/cos(declination * PI / 180.);
         update_tdi_parameters();
         break;
       case CANCEL:
         break;
    }
    XtUnmanageChild(w);
}

/* Callback function for the feedback mode toggle buttons */

void feedbackCB(w,client_data,call_data)
    Widget w;
    int client_data;
    XmAnyCallbackStruct *call_data;
{
    int n;

    /* identify the button code*/
    n=client_data;
       
    /* assign a feedback code */

     switch (n) {

       /* Control through the drive tracking system */    
       case 0: 
         feedbackmode=TRACKING;
         break; 
          
       /* Control through the adaptive optics adapter */
       case 1: 
         feedbackmode=ADAPTIVE;
         break;
                   
     } 
}

/* Callback function for the 9 grid buttons */

void buttonCB(w,client_data,call_event)
   Widget w;
   int client_data;
   XEvent *call_event;
{
   int n;
   int et = call_event->type;
   int dn,up;
    
   /* identify the button number n = 3*x+y */
   n=client_data;

   /* identify the action */
   
   dn = et == ButtonPress;
   up = et == ButtonRelease;
       
   /* respond to the request */

   if(dn){

     switch (n) {
    
       /* NoOp */
       case 0:
       break;

       /* E */
       case 1: 
       if(feedbackmode==TRACKING)
       {
         trackrelay=EAST;
       }
       else if(feedbackmode==ADAPTIVE)
       {
         ao_x=ao_x+1;
         if(ao_x>4095)
         {
           ao_x=4095;
         }
       }
       break;

       /* NoOp */
       case 2: 
       break;

       /* N */
       case 3: 
       if(feedbackmode==TRACKING)
       {
         trackrelay=NORTH;
       }
       else if(feedbackmode==ADAPTIVE)
       {
         ao_y=ao_y+1;
         if(ao_x>4095)
         {
           ao_y=4095;
         }
       }


       break;

       /* NoOp */
       case 4:
       break;

       /* S */
       case 5: 
       if(feedbackmode==TRACKING)
       {
         trackrelay=SOUTH;
       }
       else if(feedbackmode==ADAPTIVE)
       {
         ao_y=ao_y-1;
         if(ao_y<0)
         {
           ao_y=0;
         }
       }


       break;

       /* NoOp */
       case 6:
       break;

       /* W */
       case 7:
       if(feedbackmode==TRACKING)
       {
         trackrelay=WEST;
       }
       else if(feedbackmode==ADAPTIVE)
       {
         ao_x=ao_x-1;
         if(ao_x<0)
         {
           ao_x=0;
         }
       }

       break;

       /* NoOp */
       case 8: 
       break;    

     }
     
     if(feedbackmode==TRACKING)
     {
       ccd_relay_on();
     }
     else if(feedbackmode==ADAPTIVE)
     {
       ao_tip_tilt(ao_x,ao_y);
     }
     
   }
   
   if(up)
   {
     if(feedbackmode==TRACKING)
     {
       
       /* wait for the telescope to move before accepting another cycle */

       ccd_relay_off();
         
       usleep(100000.);
     }
     else if(feedbackmode==ADAPTIVE)
     {
      
       /* wait for the camera to make the step requested by the down press */
       
       usleep(5000.);
     }  
   }
        
    
}


/* Callback function for the CCD command buttons */

void commandCB(w,client_data,call_data)
   Widget w;
   int client_data;
   XmAnyCallbackStruct *call_data;
{
    
   int n, cmd_number;

   /* identify the button */
   cmd_number=client_data;

   /* respond to the request */
   n=cmd_number;

   switch (n) {

      /* Expose image CCD*/
      case 0:
        is_dark=FALSE;
        ccd=IMAGE;
        start_ccd_exposure();
        break;
        
      /* Exposure image CCD dark frame */
      case 1: 
        is_dark=TRUE;
        ccd=IMAGE;
        start_ccd_exposure();      
        break;        
        
      /* Exposure track CCD */
      case 2: 
        is_dark=FALSE;
        ccd=TRACK;
        start_ccd_exposure();
        break;

      /* Interrupt the ccd exposure cycle */
      case 3:
        if(exposure_in_progress==TRUE)
        { 
          interrupt=TRUE;
        }
        else
        {
          interrupt=FALSE;
        }
        break;

      /* Increment the filter wheel */
      case 4: 
        increment_filter();
        set_filter();
        update_filter();
        break;
   }
}


/* Update tracking ccd exposure time in the window */

void update_track_exposure()
{
    XmString xmstr;
    Arg al[10];
    int ac;
    sprintf(buf,"%4.1lf",track_exposure_time);
    xmstr = XmStringCreateLtoR(buf,char_set);
    ac=0;
    XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
    XtSetValues(show_info[0],al,ac);   
    XmStringFree(xmstr);
}    

/* Update imaging ccd exposure time in the window */

void update_image_exposure()
{
    XmString xmstr;
    Arg al[10];
    int ac;
    sprintf(buf,"%4.1lf",image_exposure_time);
    xmstr = XmStringCreateLtoR(buf,char_set);
    ac=0;
    XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
    XtSetValues(show_info[1],al,ac);   
    XmStringFree(xmstr);
}  


/* Update ccd temperature in the window */

void update_ccd_temperature()
{
    XmString xmstr;
    Arg al[10];
    int ac;
    sprintf(buf,"%3.1lf [%3.1lf (%3.0lf )]",
      ccd_temperature,ccd_setpoint_temperature,ccd_cooler_power);
    xmstr = XmStringCreateLtoR(buf,char_set);
    ac=0;
    XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
    XtSetValues(show_info[2],al,ac);   
    XmStringFree(xmstr);
}        


/* Update tdi parameters in the window */

void update_tdi_parameters()
{
    XmString xmstr;
    Arg al[10];
    int ac;
    sprintf(buf,"%6d",tdi_delay);
    xmstr = XmStringCreateLtoR(buf,char_set);
    ac=0;
    XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
    XtSetValues(show_info[3],al,ac);   
    XmStringFree(xmstr);
}        


/* Update filter in the window */

void update_filter()
{
    XmString xmstr;
    Arg al[10];
    int ac; 
    int i, imax;
    
    imax = filtermax -1;
    i = filter - 1;
    
    if ( ((i<0) || (i > imax))  || (camera_type == NO_CAMERA) )  
    {  
      xmstr = XmStringCreateLtoR(" - ",char_set);
    }
    else
    {
      xmstr = XmStringCreateLtoR(cfw_filter_name[i],char_set);
    }
    
    ac=0;
    XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
    XtSetValues(show_info[4],al,ac);   
    XmStringFree(xmstr);
} 


/* Update status in the window */

void update_status()
{
    XmString xmstr;
    Arg al[10];
    int ac;

    xmstr = XmStringCreateLtoR(status_buf,char_set);
    ac=0;
    XtSetArg(al[ac],XmNlabelString,xmstr); ac++;
    XtSetValues(show_info[5],al,ac);   
    XmStringFree(xmstr);
}   

/* Update all display buffers */

void update_all()
{
  update_status();
  update_ccd_temperature();
  if (startup_status_flag == 1)
  {
    update_tdi_parameters();
    update_filter();
    update_image_exposure();
    update_track_exposure();
    startup_status_flag = 0;
  }  
}   



/* User input of tracking ccd exposure time */

void input_track_time() 
{
   XtManageChild(track_exposure_dialog);
}


/* User input of image ccd exposure time */

void input_image_time() 
{
   XtManageChild(image_exposure_dialog);
}

/* User input of ccd temperature setpoint */

void input_ccd_setpoint() 
{
   XtManageChild(temperature_dialog);
}

/* SBIG CCD routines begin here */

char *get_error(PAR_ERROR error)
{
    char *s = "";

    switch(error) {
    case CE_NO_ERROR: return s;
    case CE_CAMERA_NOT_FOUND: s = "CE_CAMERA_NOT_FOUND"; break;
    case CE_EXPOSURE_IN_PROGRESS: s = "CE_EXPOSURE_IN_PROGRESS"; break;
    case CE_NO_EXPOSURE_IN_PROGRESS: s = "CE_NO_EXPOSURE_IN_PROGRESS"; break;
    case CE_UNKNOWN_COMMAND: s = "CE_UNKNOWN_COMMAND"; break;
    case CE_BAD_CAMERA_COMMAND: s = "CE_BAD_CAMERA_COMMAND"; break;
    case CE_BAD_PARAMETER: s = "CE_BAD_PARAMETER"; break;
    case CE_TX_TIMEOUT: s = "CE_TX_TIMEOUT"; break;
    case CE_RX_TIMEOUT: s = "CE_RX_TIMEOUT"; break;
    case CE_NAK_RECEIVED: s = "CE_NAK_RESBIGIVED"; break;
    case CE_CAN_RECEIVED: s = "CE_CAN_RESBIGIVED"; break;
    case CE_UNKNOWN_RESPONSE: s = "CE_UNKNOWN_RESPONSE"; break;
    case CE_BAD_LENGTH: s = "CE_BAD_LENGTH"; break;
    case CE_AD_TIMEOUT: s = "CE_AD_TIMEOUT"; break;
    case CE_KBD_ESC: s = "CE_KBD_ESC"; break;
    case CE_CHECKSUM_ERROR: s = "CE_CHECKSUM_ERROR"; break;
    case CE_EEPROM_ERROR: s = "CE_EEPROM_ERROR"; break;
    case CE_SHUTTER_ERROR: s = "CE_SHUTTER_ERROR"; break;
    case CE_UNKNOWN_CAMERA: s = "CE_UNKNOWN_CAMERA"; break;
    case CE_DRIVER_NOT_FOUND: s = "CE_DRIVER_NOT_FOUND"; break;
    case CE_DRIVER_NOT_OPEN: s = "CE_DRIVER_NOT_OPEN"; break;
    case CE_DRIVER_NOT_CLOSED: s = "CE_DRIVER_NOT_CLOSED"; break;
    case CE_SHARE_ERROR: s = "CE_SHARE_ERROR"; break;
    case CE_TCE_NOT_FOUND: s = "CE_TCE_NOT_FOUND"; break;
    case CE_AO_ERROR: s = "CE_AO_ERROR"; break;
    case CE_ECP_ERROR: s = "CE_ECP_ERROR"; break;
    case CE_MEMORY_ERROR: s = "CE_MEMORY_ERROR"; break;
    case CE_DEVICE_NOT_FOUND: s = "CE_DEVICE_NOT_FOUND"; break;
    case CE_DEVICE_NOT_OPEN: s = "CE_DEVICE_NOT_OPEN"; break;
    case CE_DEVICE_NOT_CLOSED: s = "CE_DEVICE_NOT_CLOSED"; break;
    case CE_DEVICE_NOT_IMPLEMENTED: s = "CE_DEVICE_NOT_IMPLEMENTED"; break;
    case CE_DEVICE_DISABLED: s = "CE_DEVICE_DISABLED"; break;
    case CE_OS_ERROR: s = "CE_OS_ERROR"; break;
    case CE_SOCK_ERROR: s = "CE_SOCK_ERROR"; break;
    case CE_SERVER_NOT_FOUND: s = "CE_SERVER_NOT_FOUND"; break;
    case CE_CFW_ERROR: s = "CE_CFW_ERROR"; break;
    case CE_NEXT_ERROR: s = "CE_NEXT_ERROR"; break;
    }
    return s;
}

void show_error(char *s, int err)
{
/* Uncomment for a longer diagnostic         */
/*    printf("ERROR: %s: %d\n", s, err);   */

    sprintf(status_buf,"Error %s %d",s,err);
}


/* Write a FITS primary array with a 2-D image */
/*   and a header with optional keywords       */
/*    INSTRUME  CCD Camera             */
/*    DATE-OBS  ISO 8601 Date and Time */
/*    EXPTIME   Exposure time          */
/*    HISTORY                          */
/*    FILTER    Filter description     */



void write_fits(char *filename, int w, int h, unsigned short *data)
{
    fitsfile *fptr;       /* pointer to the FITS file, defined in fitsio.h */
    int status;
    long  fpixel, nelements;
    double exposure;

    /* Initialize FITS image parameters */

    int bitpix   =  USHORT_IMG;       /* 16-bit unsigned short pixel values */
    long naxis    =   2;              /* 2-dimensional image                */    
    long naxes[2] = { 256,256 };      /* default image 256 wide by 256 rows */

    /* Header information */

    char camera_type_buf[16];       /* buffer for camera type string */
    char camera_mode_buf[16];       /* buffer for camera mode string */
    char camera_note_buf[16];       /* buffer for camera note string */
    char camera_filter_buf[24];     /* buffer for camera filter string */
    char camera_time_buf[24];       /* buffer for camera time string */
    time_t timebuf; 
    struct tm *ptm;

    /* Set the actual width and height of the image */

    naxes[0] = w;
    naxes[1] = h;

    /* Delete old FITS file if it already exists */  

    remove(filename); 
     
    /* Me must initialize status before calling fitsio routines */           

    status = 0;  
    
    /* Create a new FITS file and show error message if one occurs */

    if (fits_create_file(&fptr, filename, &status)) 
         show_cfitsio_error( status );           

    /* Write the required keywords for the primary array image.       */
    /* Since bitpix = USHORT_IMG, this will cause cfitsio to create   */
    /* a FITS image with BITPIX = 16 (signed short integers) with     */
    /* BSCALE = 1.0 and BZERO = 32768.  This is the convention that   */
    /* FITS uses to store unsigned integers.  Note that the BSCALE    */
    /* and BZERO keywords will be automatically written by cfitsio    */
    /* in this case.                                                */

    if ( fits_create_img(fptr,  bitpix, naxis, naxes, &status) )
        show_cfitsio_error( status );          

    fpixel = 1;                               /* first pixel to write      */
    nelements = naxes[0] * naxes[1];          /* number of pixels to write */

    /* Write the array of unsigned integers to the FITS file */
    
    if ( fits_write_img(fptr, TUSHORT, fpixel, nelements, data, &status) )
        show_cfitsio_error( status );
        
    /* Write optional keywords to the header */
    /* Note that the ADDRESS of the value is passed in the routine */


    switch (camera_type) {
	case ST7_CAMERA:
            sprintf(camera_type_buf,"SBIG ST7");
	    break;
	case ST8_CAMERA:
            sprintf(camera_type_buf,"SBIG ST8");
	    break;
	case ST9_CAMERA:
            sprintf(camera_type_buf,"SBIG ST9");
	    break;         
	case ST10_CAMERA:
            sprintf(camera_type_buf,"SBIG ST10");
	    break; 
	case STL_CAMERA:
            sprintf(camera_type_buf,"SBIG STL");
	    break; 
	case ST402_CAMERA:
            sprintf(camera_type_buf,"SBIG ST402");
	    break; 

	default:
            sprintf(camera_type_buf,"SBIG CCD");        
    }
    
    if( readout_ccd==TRACK )
    {   
        sprintf(camera_mode_buf,"Track CCD");
        exposure = track_exposure_time;
    }
    else
    {
        sprintf(camera_mode_buf,"Image CCD");  
        exposure = image_exposure_time;
    }

    if ( is_dark )
    {
      sprintf(camera_note_buf, "Dark Frame");
    }
    else
    {
      sprintf(camera_note_buf, "Light Frame");
    }
     
    time( &timebuf );
    ptm = gmtime( &timebuf );
    
    /* ptm->tm_mon is from 0 to 11 and must be incremented by 1 */
    
    ++ptm->tm_mon;
    
    sprintf(camera_time_buf, "%02d:%02d:%02d %4d-%02d-%02d", 
      ptm->tm_hour, ptm->tm_min, ptm->tm_sec, 
      1900 + ptm->tm_year, ptm->tm_mon, ptm->tm_mday );

    sprintf(camera_filter_buf, cfw_filter_name[filter-1]);

    if ( fits_update_key_str(fptr, "INSTRUME", 
         camera_type_buf, camera_mode_buf, &status) )
         show_cfitsio_error( status ); 
         
    if ( fits_update_key_dbl(fptr, "EXPTIME", exposure, -2,
         "exposure time (seconds)", &status) )
         show_cfitsio_error( status ); 
         
    if ( fits_update_key_str(fptr, "DATE", 
         camera_time_buf, NULL, &status) )
         show_cfitsio_error( status );         
         
    if ( fits_update_key_str(fptr, "FILTER", 
         camera_filter_buf,"Filter", &status) )
         show_cfitsio_error( status );  
            
    if ( fits_update_key_str(fptr, "IMAGETYP", 
         camera_note_buf, NULL, &status) )
         show_cfitsio_error( status );                   
          
 
    /* Close the file */             

    if ( fits_close_file(fptr, &status) )              
         show_cfitsio_error( status );           

    return;
}

    
    
void establish_link(int camera_device)
{
    int err;
    OpenDeviceParams odp;
    EstablishLinkParams  in;
    EstablishLinkResults  out;

    switch (camera_device) {
	case DEV_LPT1:
	    odp.deviceType = DEV_LPT1;
            odp.lptBaseAddress = 0x378;
	    break;
	case DEV_LPT2:
	    odp.deviceType = DEV_LPT2;
            odp.lptBaseAddress = 0x3bc;
	    break;
	case DEV_LPT3:
	    odp.deviceType = DEV_LPT3;
            odp.lptBaseAddress = 0x278;           
	    break;
	case DEV_USB:
	    odp.deviceType = DEV_USB;
	    break;
	case DEV_ETH:
	    odp.deviceType = DEV_ETH;
            odp.ipAddress = camera_ipaddress;   
	    break;
	case DEV_USB1:
	    odp.deviceType = DEV_USB1;
	    break;
	case DEV_USB2:
	    odp.deviceType = DEV_USB2;
	    break;
	case DEV_USB3:
	    odp.deviceType = DEV_USB3;
	    break;
 	case DEV_USB4:
	    odp.deviceType = DEV_USB4;
	    break;           
	default:
	    fprintf(stderr, "Unsupported camera device %d \n",
	      camera_device);
	    exit(1);
    }

    err = SBIGUnivDrvCommand(CC_OPEN_DEVICE, &odp, NULL);
    if (err != CE_NO_ERROR)
    {
       GetErrorStringParams  gesp;
       GetErrorStringResults  gesr;
       gesp.errorNo = err;
       SBIGUnivDrvCommand(CC_GET_ERROR_STRING, &gesp, &gesr);
/*
       fprintf(stderr,"CC_OPEN_DEVICE command failed in xmccd: %s\n", &gesr);
*/
       fprintf(stderr,"CC_OPEN_DEVICE command failed in xmccd\n");
    }

    err = SBIGUnivDrvCommand(CC_ESTABLISH_LINK, &in, &out);
    if ( err == CE_NO_ERROR ) 
    {
       camera_type = out.cameraType;
    } 
    else
    {
       camera_type = NO_CAMERA;
       GetErrorStringParams  gesp;
       GetErrorStringResults  gesr;
       gesp.errorNo = err;
       SBIGUnivDrvCommand(CC_GET_ERROR_STRING, &gesp, &gesr);
/*
       fprintf(stderr,"CC_ESTABLISH_LINK command failed in xmccd:\n  %s\n", &gesr);
*/
       fprintf(stderr,"CC_ESTABLISH_LINK command failed in xmccd\n");

    }

}

void  get_info(int type)
{
    int  result;
    GetCCDInfoParams  in;

    in.request = type;
    if (type == 0)
	result = SBIGUnivDrvCommand(CC_GET_CCD_INFO, &in, &info0);
    else if (type == 1)
	result = SBIGUnivDrvCommand(CC_GET_CCD_INFO, &in, &info1);
    else if (type == 2)
	result = SBIGUnivDrvCommand(CC_GET_CCD_INFO, &in, &info2);
    else if (type == 3)
	result = SBIGUnivDrvCommand(CC_GET_CCD_INFO, &in, &info3);
    else {
	return;
    }
    if (result != 0) 
    {
       return;
    }

    if (type == 0 || type == 1) 
    {
       GetCCDInfoResults0  *info = (type == 0? &info0 : &info1);
    } 
}

void  get_ccd_info(void)
{
    if (camera_type == NO_CAMERA) 
    {
	return;
    }
    get_info(0);
    get_info(1);
    get_info(2);

}


/* Set panel message buffer to show the status of the ccd system       */
/* Status values have been previously read from the CCD and processed  */
/* Check status == CS_INTEGRATION_COMPLETE to flag ready to readout    */

static void show_ccd_status(int imaging_ccd_status, int tracking_ccd_status)
{
    if (imaging_ccd_status == 2)
	image_in_progress = 1;
    else if (imaging_ccd_status == 3)
	image_in_progress = 2;
    else
	image_in_progress = 0;
    if (tracking_ccd_status == 2)
	track_in_progress = 1;
    else if (tracking_ccd_status == 3)
	track_in_progress = 2;
    else
	track_in_progress = 0;

    if (image_in_progress != 0) {
	if (image_in_progress == 1)
        {
            if(image_mode==SINGLE)
            {
              if(is_dark)
              {
                sprintf(status_buf,"Single dark image");
              }
              else
              {
                sprintf(status_buf,"Single image");
              }  
            }
            else if(image_mode==FOCUS)
            {
              if(is_dark)
              {
                sprintf(status_buf,"Focus dark image");
              }
              else
              {
                sprintf(status_buf,"Focus image");
              }               
            }
            else if(image_mode==MULTI)
            {
              if(is_dark)
              {
                sprintf(status_buf,"Multiple dark images");
              }
              else
              {
                sprintf(status_buf,"Multiple images");
              } 
            }
            else if(image_mode==TDI)
            {
              if(is_dark)
              {
                sprintf(status_buf,"TDI dark image");
              }
              else
              {
                sprintf(status_buf,"TDI image");
              } 
            }
            else 
            {
              sprintf(status_buf,"Image");
            }             
            update_all();
	}
        else
	{
            sprintf(status_buf,"Image readout");
            update_all();
        }   
    }
    if (track_in_progress != 0) {
	if (track_in_progress == 1)
	{
            if(image_mode==SINGLE)
            {
              if(is_dark)
              {
                sprintf(status_buf,"Single dark track");
              }
              else
              {
                sprintf(status_buf,"Single track");
              }  
            }
            else if(image_mode==FOCUS)
            {
              if(is_dark)
              {
                sprintf(status_buf,"Focus dark track");
              }
              else
              {
                sprintf(status_buf,"Focus track");
              }               
            }
            else if(image_mode==MULTI)
            {
              if(is_dark)
              {
                sprintf(status_buf,"Multiple dark tracks");
              }
              else
              {
                sprintf(status_buf,"Multiple tracks");
              } 
            }
            else 
            {
              sprintf(status_buf,"Track");
            }
            update_all();
	}
        else
	{
            sprintf(status_buf,"Track readout");
            update_all();
        }
    }
    
    if ((image_in_progress == 0) & (track_in_progress == 0))
    {  
     if (camera_type != NO_CAMERA)
      {
        sprintf(status_buf,"Camera ready");
      }
      update_all();
    } 
}


static short ccd_temp[4096] = {
10000,2134,1950,1842,1766,1707,1659,1618,1582,1551,1523,1498,
1475,1453,1434,1415,1398,1382,1367,1353,1339,1326,1314,1302,
1290,1279,1269,1259,1249,1240,1231,1222,1214,1205,1197,1190,
1182,1175,1168,1161,1154,1147,1141,1135,1129,1122,1117,1111,
1105,1100,1094,1089,1084,1079,1074,1069,1064,1059,1054,1050,
1045,1041,1036,1032,1028,1024,1020,1016,1012,1008,1004,1000,
996,993,989,985,982,978,975,971,968,964,961,958,
955,951,948,945,942,939,936,933,930,927,924,921,
918,916,913,910,907,905,902,899,897,894,892,889,
886,884,881,879,877,874,872,869,867,865,862,860,
858,855,853,851,849,847,844,842,840,838,836,834,
832,830,828,826,824,822,820,818,816,814,812,810,
808,806,804,802,800,798,797,795,793,791,789,788,
786,784,782,781,779,777,775,774,772,770,769,767,
765,764,762,761,759,757,756,754,753,751,749,748,
746,745,743,742,740,739,737,736,734,733,731,730,
728,727,726,724,723,721,720,718,717,716,714,713,
712,710,709,707,706,705,703,702,701,700,698,697,
696,694,693,692,690,689,688,687,685,684,683,682,
680,679,678,677,676,674,673,672,671,670,668,667,
666,665,664,663,661,660,659,658,657,656,655,653,
652,651,650,649,648,647,646,645,643,642,641,640,
639,638,637,636,635,634,633,632,631,630,629,628,
627,626,625,623,622,621,620,619,618,617,616,615,
614,613,612,611,611,610,609,608,607,606,605,604,
603,602,601,600,599,598,597,596,595,594,593,592,
592,591,590,589,588,587,586,585,584,583,583,582,
581,580,579,578,577,576,575,575,574,573,572,571,
570,569,569,568,567,566,565,564,563,563,562,561,
560,559,558,558,557,556,555,554,554,553,552,551,
550,549,549,548,547,546,545,545,544,543,542,542,
541,540,539,538,538,537,536,535,535,534,533,532,
531,531,530,529,528,528,527,526,525,525,524,523,
522,522,521,520,520,519,518,517,517,516,515,514,
514,513,512,512,511,510,509,509,508,507,507,506,
505,504,504,503,502,502,501,500,500,499,498,498,
497,496,495,495,494,493,493,492,491,491,490,489,
489,488,487,487,486,485,485,484,483,483,482,481,
481,480,479,479,478,478,477,476,476,475,474,474,
473,472,472,471,470,470,469,469,468,467,467,466,
465,465,464,464,463,462,462,461,460,460,459,459,
458,457,457,456,456,455,454,454,453,453,452,451,
451,450,450,449,448,448,447,447,446,445,445,444,
444,443,442,442,441,441,440,440,439,438,438,437,
437,436,436,435,434,434,433,433,432,432,431,430,
430,429,429,428,428,427,426,426,425,425,424,424,
423,423,422,421,421,420,420,419,419,418,418,417,
417,416,415,415,414,414,413,413,412,412,411,411,
410,410,409,408,408,407,407,406,406,405,405,404,
404,403,403,402,402,401,401,400,399,399,398,398,
397,397,396,396,395,395,394,394,393,393,392,392,
391,391,390,390,389,389,388,388,387,387,386,386,
385,385,384,384,383,383,382,382,381,381,380,380,
379,379,378,378,377,377,376,376,375,375,374,374,
373,373,372,372,371,371,370,370,370,369,369,368,
368,367,367,366,366,365,365,364,364,363,363,362,
362,361,361,360,360,360,359,359,358,358,357,357,
356,356,355,355,354,354,353,353,353,352,352,351,
351,350,350,349,349,348,348,348,347,347,346,346,
345,345,344,344,343,343,343,342,342,341,341,340,
340,339,339,339,338,338,337,337,336,336,335,335,
335,334,334,333,333,332,332,331,331,331,330,330,
329,329,328,328,328,327,327,326,326,325,325,325,
324,324,323,323,322,322,322,321,321,320,320,319,
319,319,318,318,317,317,316,316,316,315,315,314,
314,314,313,313,312,312,311,311,311,310,310,309,
309,309,308,308,307,307,306,306,306,305,305,304,
304,304,303,303,302,302,302,301,301,300,300,299,
299,299,298,298,297,297,297,296,296,295,295,295,
294,294,293,293,293,292,292,291,291,291,290,290,
289,289,289,288,288,287,287,287,286,286,286,285,
285,284,284,284,283,283,282,282,282,281,281,280,
280,280,279,279,278,278,278,277,277,277,276,276,
275,275,275,274,274,273,273,273,272,272,272,271,
271,270,270,270,269,269,269,268,268,267,267,267,
266,266,266,265,265,264,264,264,263,263,263,262,
262,261,261,261,260,260,260,259,259,258,258,258,
257,257,257,256,256,255,255,255,254,254,254,253,
253,253,252,252,251,251,251,250,250,250,249,249,
248,248,248,247,247,247,246,246,246,245,245,245,
244,244,243,243,243,242,242,242,241,241,241,240,
240,239,239,239,238,238,238,237,237,237,236,236,
236,235,235,235,234,234,233,233,233,232,232,232,
231,231,231,230,230,230,229,229,229,228,228,227,
227,227,226,226,226,225,225,225,224,224,224,223,
223,223,222,222,222,221,221,221,220,220,220,219,
219,218,218,218,217,217,217,216,216,216,215,215,
215,214,214,214,213,213,213,212,212,212,211,211,
211,210,210,210,209,209,209,208,208,208,207,207,
207,206,206,206,205,205,205,204,204,204,203,203,
203,202,202,202,201,201,201,200,200,200,199,199,
199,198,198,198,197,197,197,196,196,196,195,195,
195,194,194,194,193,193,193,192,192,192,191,191,
191,190,190,190,189,189,189,188,188,188,188,187,
187,187,186,186,186,185,185,185,184,184,184,183,
183,183,182,182,182,181,181,181,180,180,180,179,
179,179,178,178,178,178,177,177,177,176,176,176,
175,175,175,174,174,174,173,173,173,172,172,172,
171,171,171,171,170,170,170,169,169,169,168,168,
168,167,167,167,166,166,166,166,165,165,165,164,
164,164,163,163,163,162,162,162,161,161,161,161,
160,160,160,159,159,159,158,158,158,157,157,157,
157,156,156,156,155,155,155,154,154,154,153,153,
153,153,152,152,152,151,151,151,150,150,150,149,
149,149,149,148,148,148,147,147,147,146,146,146,
145,145,145,145,144,144,144,143,143,143,142,142,
142,142,141,141,141,140,140,140,139,139,139,139,
138,138,138,137,137,137,136,136,136,136,135,135,
135,134,134,134,133,133,133,133,132,132,132,131,
131,131,130,130,130,130,129,129,129,128,128,128,
128,127,127,127,126,126,126,125,125,125,125,124,
124,124,123,123,123,122,122,122,122,121,121,121,
120,120,120,120,119,119,119,118,118,118,118,117,
117,117,116,116,116,115,115,115,115,114,114,114,
113,113,113,113,112,112,112,111,111,111,111,110,
110,110,109,109,109,109,108,108,108,107,107,107,
106,106,106,106,105,105,105,104,104,104,104,103,
103,103,102,102,102,102,101,101,101,100,100,100,
100,99,99,99,98,98,98,98,97,97,97,96,
96,96,96,95,95,95,94,94,94,94,93,93,
93,92,92,92,92,91,91,91,90,90,90,90,
89,89,89,89,88,88,88,87,87,87,87,86,
86,86,85,85,85,85,84,84,84,83,83,83,
83,82,82,82,81,81,81,81,80,80,80,80,
79,79,79,78,78,78,78,77,77,77,76,76,
76,76,75,75,75,75,74,74,74,73,73,73,
73,72,72,72,71,71,71,71,70,70,70,70,
69,69,69,68,68,68,68,67,67,67,66,66,
66,66,65,65,65,65,64,64,64,63,63,63,
63,62,62,62,62,61,61,61,60,60,60,60,
59,59,59,58,58,58,58,57,57,57,57,56,
56,56,55,55,55,55,54,54,54,54,53,53,
53,52,52,52,52,51,51,51,51,50,50,50,
49,49,49,49,48,48,48,48,47,47,47,46,
46,46,46,45,45,45,45,44,44,44,44,43,
43,43,42,42,42,42,41,41,41,41,40,40,
40,39,39,39,39,38,38,38,38,37,37,37,
37,36,36,36,35,35,35,35,34,34,34,34,
33,33,33,32,32,32,32,31,31,31,31,30,
30,30,30,29,29,29,28,28,28,28,27,27,
27,27,26,26,26,26,25,25,25,24,24,24,
24,23,23,23,23,22,22,22,22,21,21,21,
20,20,20,20,19,19,19,19,18,18,18,18,
17,17,17,16,16,16,16,15,15,15,15,14,
14,14,14,13,13,13,12,12,12,12,11,11,
11,11,10,10,10,10,9,9,9,9,8,8,
8,7,7,7,7,6,6,6,6,5,5,5,
5,4,4,4,3,3,3,3,2,2,2,2,
1,1,1,1,0,0,0,0,0,0,0,-1,
-1,-1,-1,-2,-2,-2,-2,-3,-3,-3,-3,-4,
-4,-4,-4,-5,-5,-5,-5,-6,-6,-6,-7,-7,
-7,-7,-8,-8,-8,-8,-9,-9,-9,-9,-10,-10,
-10,-10,-11,-11,-11,-12,-12,-12,-12,-13,-13,-13,
-13,-14,-14,-14,-14,-15,-15,-15,-15,-16,-16,-16,
-16,-17,-17,-17,-18,-18,-18,-18,-19,-19,-19,-19,
-20,-20,-20,-20,-21,-21,-21,-21,-22,-22,-22,-22,
-23,-23,-23,-24,-24,-24,-24,-25,-25,-25,-25,-26,
-26,-26,-26,-27,-27,-27,-27,-28,-28,-28,-28,-29,
-29,-29,-30,-30,-30,-30,-31,-31,-31,-31,-32,-32,
-32,-32,-33,-33,-33,-33,-34,-34,-34,-34,-35,-35,
-35,-35,-36,-36,-36,-37,-37,-37,-37,-38,-38,-38,
-38,-39,-39,-39,-39,-40,-40,-40,-40,-41,-41,-41,
-41,-42,-42,-42,-42,-43,-43,-43,-44,-44,-44,-44,
-45,-45,-45,-45,-46,-46,-46,-46,-47,-47,-47,-47,
-48,-48,-48,-48,-49,-49,-49,-49,-50,-50,-50,-51,
-51,-51,-51,-52,-52,-52,-52,-53,-53,-53,-53,-54,
-54,-54,-54,-55,-55,-55,-55,-56,-56,-56,-56,-57,
-57,-57,-58,-58,-58,-58,-59,-59,-59,-59,-60,-60,
-60,-60,-61,-61,-61,-61,-62,-62,-62,-62,-63,-63,
-63,-63,-64,-64,-64,-65,-65,-65,-65,-66,-66,-66,
-66,-67,-67,-67,-67,-68,-68,-68,-68,-69,-69,-69,
-69,-70,-70,-70,-70,-71,-71,-71,-71,-72,-72,-72,
-73,-73,-73,-73,-74,-74,-74,-74,-75,-75,-75,-75,
-76,-76,-76,-76,-77,-77,-77,-77,-78,-78,-78,-78,
-79,-79,-79,-80,-80,-80,-80,-81,-81,-81,-81,-82,
-82,-82,-82,-83,-83,-83,-83,-84,-84,-84,-84,-85,
-85,-85,-85,-86,-86,-86,-86,-87,-87,-87,-88,-88,
-88,-88,-89,-89,-89,-89,-90,-90,-90,-90,-91,-91,
-91,-91,-92,-92,-92,-92,-93,-93,-93,-93,-94,-94,
-94,-95,-95,-95,-95,-96,-96,-96,-96,-97,-97,-97,
-97,-98,-98,-98,-98,-99,-99,-99,-99,-100,-100,-100,
-100,-101,-101,-101,-102,-102,-102,-102,-103,-103,-103,-103,
-104,-104,-104,-104,-105,-105,-105,-105,-106,-106,-106,-106,
-107,-107,-107,-108,-108,-108,-108,-109,-109,-109,-109,-110,
-110,-110,-110,-111,-111,-111,-111,-112,-112,-112,-112,-113,
-113,-113,-113,-114,-114,-114,-115,-115,-115,-115,-116,-116,
-116,-116,-117,-117,-117,-117,-118,-118,-118,-118,-119,-119,
-119,-119,-120,-120,-120,-121,-121,-121,-121,-122,-122,-122,
-122,-123,-123,-123,-123,-124,-124,-124,-124,-125,-125,-125,
-126,-126,-126,-126,-127,-127,-127,-127,-128,-128,-128,-128,
-129,-129,-129,-129,-130,-130,-130,-130,-131,-131,-131,-132,
-132,-132,-132,-133,-133,-133,-133,-134,-134,-134,-134,-135,
-135,-135,-135,-136,-136,-136,-137,-137,-137,-137,-138,-138,
-138,-138,-139,-139,-139,-139,-140,-140,-140,-140,-141,-141,
-141,-142,-142,-142,-142,-143,-143,-143,-143,-144,-144,-144,
-144,-145,-145,-145,-145,-146,-146,-146,-147,-147,-147,-147,
-148,-148,-148,-148,-149,-149,-149,-149,-150,-150,-150,-151,
-151,-151,-151,-152,-152,-152,-152,-153,-153,-153,-153,-154,
-154,-154,-154,-155,-155,-155,-156,-156,-156,-156,-157,-157,
-157,-157,-158,-158,-158,-158,-159,-159,-159,-160,-160,-160,
-160,-161,-161,-161,-161,-162,-162,-162,-162,-163,-163,-163,
-164,-164,-164,-164,-165,-165,-165,-165,-166,-166,-166,-166,
-167,-167,-167,-168,-168,-168,-168,-169,-169,-169,-169,-170,
-170,-170,-171,-171,-171,-171,-172,-172,-172,-172,-173,-173,
-173,-173,-174,-174,-174,-175,-175,-175,-175,-176,-176,-176,
-176,-177,-177,-177,-177,-178,-178,-178,-179,-179,-179,-179,
-180,-180,-180,-180,-181,-181,-181,-182,-182,-182,-182,-183,
-183,-183,-183,-184,-184,-184,-185,-185,-185,-185,-186,-186,
-186,-186,-187,-187,-187,-188,-188,-188,-188,-189,-189,-189,
-189,-190,-190,-190,-191,-191,-191,-191,-192,-192,-192,-192,
-193,-193,-193,-194,-194,-194,-194,-195,-195,-195,-195,-196,
-196,-196,-197,-197,-197,-197,-198,-198,-198,-198,-199,-199,
-199,-200,-200,-200,-200,-201,-201,-201,-201,-202,-202,-202,
-203,-203,-203,-203,-204,-204,-204,-205,-205,-205,-205,-206,
-206,-206,-206,-207,-207,-207,-208,-208,-208,-208,-209,-209,
-209,-209,-210,-210,-210,-211,-211,-211,-211,-212,-212,-212,
-213,-213,-213,-213,-214,-214,-214,-214,-215,-215,-215,-216,
-216,-216,-216,-217,-217,-217,-218,-218,-218,-218,-219,-219,
-219,-220,-220,-220,-220,-221,-221,-221,-221,-222,-222,-222,
-223,-223,-223,-223,-224,-224,-224,-225,-225,-225,-225,-226,
-226,-226,-227,-227,-227,-227,-228,-228,-228,-229,-229,-229,
-229,-230,-230,-230,-231,-231,-231,-231,-232,-232,-232,-233,
-233,-233,-233,-234,-234,-234,-235,-235,-235,-235,-236,-236,
-236,-236,-237,-237,-237,-238,-238,-238,-238,-239,-239,-239,
-240,-240,-240,-241,-241,-241,-241,-242,-242,-242,-243,-243,
-243,-243,-244,-244,-244,-245,-245,-245,-245,-246,-246,-246,
-247,-247,-247,-247,-248,-248,-248,-249,-249,-249,-249,-250,
-250,-250,-251,-251,-251,-251,-252,-252,-252,-253,-253,-253,
-254,-254,-254,-254,-255,-255,-255,-256,-256,-256,-256,-257,
-257,-257,-258,-258,-258,-258,-259,-259,-259,-260,-260,-260,
-261,-261,-261,-261,-262,-262,-262,-263,-263,-263,-263,-264,
-264,-264,-265,-265,-265,-266,-266,-266,-266,-267,-267,-267,
-268,-268,-268,-269,-269,-269,-269,-270,-270,-270,-271,-271,
-271,-272,-272,-272,-272,-273,-273,-273,-274,-274,-274,-275,
-275,-275,-275,-276,-276,-276,-277,-277,-277,-278,-278,-278,
-278,-279,-279,-279,-280,-280,-280,-281,-281,-281,-281,-282,
-282,-282,-283,-283,-283,-284,-284,-284,-284,-285,-285,-285,
-286,-286,-286,-287,-287,-287,-288,-288,-288,-288,-289,-289,
-289,-290,-290,-290,-291,-291,-291,-292,-292,-292,-292,-293,
-293,-293,-294,-294,-294,-295,-295,-295,-296,-296,-296,-296,
-297,-297,-297,-298,-298,-298,-299,-299,-299,-300,-300,-300,
-300,-301,-301,-301,-302,-302,-302,-303,-303,-303,-304,-304,
-304,-305,-305,-305,-305,-306,-306,-306,-307,-307,-307,-308,
-308,-308,-309,-309,-309,-310,-310,-310,-311,-311,-311,-311,
-312,-312,-312,-313,-313,-313,-314,-314,-314,-315,-315,-315,
-316,-316,-316,-317,-317,-317,-318,-318,-318,-318,-319,-319,
-319,-320,-320,-320,-321,-321,-321,-322,-322,-322,-323,-323,
-323,-324,-324,-324,-325,-325,-325,-326,-326,-326,-327,-327,
-327,-328,-328,-328,-329,-329,-329,-329,-330,-330,-330,-331,
-331,-331,-332,-332,-332,-333,-333,-333,-334,-334,-334,-335,
-335,-335,-336,-336,-336,-337,-337,-337,-338,-338,-338,-339,
-339,-339,-340,-340,-340,-341,-341,-341,-342,-342,-342,-343,
-343,-343,-344,-344,-344,-345,-345,-345,-346,-346,-346,-347,
-347,-347,-348,-348,-348,-349,-349,-349,-350,-350,-350,-351,
-351,-351,-352,-352,-353,-353,-353,-354,-354,-354,-355,-355,
-355,-356,-356,-356,-357,-357,-357,-358,-358,-358,-359,-359,
-359,-360,-360,-360,-361,-361,-361,-362,-362,-362,-363,-363,
-364,-364,-364,-365,-365,-365,-366,-366,-366,-367,-367,-367,
-368,-368,-368,-369,-369,-369,-370,-370,-371,-371,-371,-372,
-372,-372,-373,-373,-373,-374,-374,-374,-375,-375,-375,-376,
-376,-377,-377,-377,-378,-378,-378,-379,-379,-379,-380,-380,
-380,-381,-381,-382,-382,-382,-383,-383,-383,-384,-384,-384,
-385,-385,-386,-386,-386,-387,-387,-387,-388,-388,-388,-389,
-389,-390,-390,-390,-391,-391,-391,-392,-392,-392,-393,-393,
-394,-394,-394,-395,-395,-395,-396,-396,-397,-397,-397,-398,
-398,-398,-399,-399,-400,-400,-400,-401,-401,-401,-402,-402,
-403,-403,-403,-404,-404,-404,-405,-405,-406,-406,-406,-407,
-407,-407,-408,-408,-409,-409,-409,-410,-410,-410,-411,-411,
-412,-412,-412,-413,-413,-414,-414,-414,-415,-415,-415,-416,
-416,-417,-417,-417,-418,-418,-419,-419,-419,-420,-420,-421,
-421,-421,-422,-422,-422,-423,-423,-424,-424,-424,-425,-425,
-426,-426,-426,-427,-427,-428,-428,-428,-429,-429,-430,-430,
-430,-431,-431,-432,-432,-432,-433,-433,-434,-434,-434,-435,
-435,-436,-436,-436,-437,-437,-438,-438,-438,-439,-439,-440,
-440,-441,-441,-441,-442,-442,-443,-443,-443,-444,-444,-445,
-445,-445,-446,-446,-447,-447,-448,-448,-448,-449,-449,-450,
-450,-450,-451,-451,-452,-452,-453,-453,-453,-454,-454,-455,
-455,-455,-456,-456,-457,-457,-458,-458,-458,-459,-459,-460,
-460,-461,-461,-461,-462,-462,-463,-463,-464,-464,-464,-465,
-465,-466,-466,-467,-467,-467,-468,-468,-469,-469,-470,-470,
-471,-471,-471,-472,-472,-473,-473,-474,-474,-475,-475,-475,
-476,-476,-477,-477,-478,-478,-479,-479,-479,-480,-480,-481,
-481,-482,-482,-483,-483,-483,-484,-484,-485,-485,-486,-486,
-487,-487,-488,-488,-488,-489,-489,-490,-490,-491,-491,-492,
-492,-493,-493,-494,-494,-494,-495,-495,-496,-496,-497,-497,
-498,-498,-499,-499,-500,-500,-501,-501,-502,-502,-502,-503,
-503,-504,-504,-505,-505,-506,-506,-507,-507,-508,-508,-509,
-509,-510,-510,-511,-511,-512,-512,-513,-513,-514,-514,-515,
-515,-516,-516,-517,-517,-517,-518,-518,-519,-519,-520,-520,
-521,-521,-522,-522,-523,-523,-524,-524,-525,-525,-526,-526,
-527,-527,-528,-529,-529,-530,-530,-531,-531,-532,-532,-533,
-533,-534,-534,-535,-535,-536,-536,-537,-537,-538,-538,-539,
-539,-540,-540,-541,-541,-542,-542,-543,-544,-544,-545,-545,
-546,-546,-547,-547,-548,-548,-549,-549,-550,-550,-551,-552,
-552,-553,-553,-554,-554,-555,-555,-556,-556,-557,-558,-558,
-559,-559,-560,-560,-561,-561,-562,-563,-563,-564,-564,-565,
-565,-566,-566,-567,-568,-568,-569,-569,-570,-570,-571,-572,
-572,-573,-573,-574,-574,-575,-576,-576,-577,-577,-578,-578,
-579,-580,-580,-581,-581,-582,-583,-583,-584,-584,-585,-586,
-586,-587,-587,-588,-589,-589,-590,-590,-591,-592,-592,-593,
-593,-594,-595,-595,-596,-596,-597,-598,-598,-599,-599,-600,
-601,-601,-602,-603,-603,-604,-604,-605,-606,-606,-607,-608,
-608,-609,-610,-610,-611,-611,-612,-613,-613,-614,-615,-615,
-616,-617,-617,-618,-619,-619,-620,-621,-621,-622,-622,-623,
-624,-624,-625,-626,-626,-627,-628,-629,-629,-630,-631,-631,
-632,-633,-633,-634,-635,-635,-636,-637,-637,-638,-639,-639,
-640,-641,-642,-642,-643,-644,-644,-645,-646,-646,-647,-648,
-649,-649,-650,-651,-651,-652,-653,-654,-654,-655,-656,-657,
-657,-658,-659,-660,-660,-661,-662,-662,-663,-664,-665,-665,
-666,-667,-668,-668,-669,-670,-671,-672,-672,-673,-674,-675,
-675,-676,-677,-678,-679,-679,-680,-681,-682,-682,-683,-684,
-685,-686,-686,-687,-688,-689,-690,-690,-691,-692,-693,-694,
-695,-695,-696,-697,-698,-699,-700,-700,-701,-702,-703,-704,
-705,-705,-706,-707,-708,-709,-710,-711,-711,-712,-713,-714,
-715,-716,-717,-718,-719,-719,-720,-721,-722,-723,-724,-725,
-726,-727,-728,-728,-729,-730,-731,-732,-733,-734,-735,-736,
-737,-738,-739,-740,-741,-742,-742,-743,-744,-745,-746,-747,
-748,-749,-750,-751,-752,-753,-754,-755,-756,-757,-758,-759,
-760,-761,-762,-763,-764,-765,-766,-767,-768,-770,-771,-772,
-773,-774,-775,-776,-777,-778,-779,-780,-781,-782,-783,-785,
-786,-787,-788,-789,-790,-791,-792,-793,-795,-796,-797,-798,
-799,-800,-801,-803,-804,-805,-806,-807,-809,-810,-811,-812,
-813,-815,-816,-817,-818,-819,-821,-822,-823,-824,-826,-827,
-828,-830,-831,-832,-833,-835,-836,-837,-839,-840,-841,-843,
-844,-845,-847,-848,-849,-851,-852,-853,-855,-856,-858,-859,
-860,-862,-863,-865,-866,-868,-869,-871,-872,-874,-875,-877,
-878,-880,-881,-883,-884,-886,-887,-889,-890,-892,-893,-895,
-897,-898,-900,-902,-903,-905,-906,-908,-910,-912,-913,-915,
-917,-918,-920,-922,-924,-925,-927,-929,-931,-933,-934,-936,
-938,-940,-942,-944,-946,-948,-949,-951,-953,-955,-957,-959,
-961,-963,-965,-967,-969,-972,-974,-976,-978,-980,-982,-984,
-987,-989,-991,-993,-996,-998,-1000,-1002,-1005,-1007,-1010,-1012,
-1014,-1017,-1019,-1022,-1024,-1027,-1029,-1032,-1034,-1037,-1040,-1042,
-1045,-1048,-1051,-1053,-1056,-1059,-1062,-1065,-1068,-1071,-1074,-1077,
-1080,-1083,-1086,-1089,-1092,-1096,-1099,-1102,-1106,-1109,-1112,-1116,
-1119,-1123,-1127,-1130,-1134,-1138,-1142,-1145,-1149,-1153,-1157,-1162,
-1166,-1170,-1174,-1179,-1183,-1188,-1192,-1197,-1202,-1206,-1211,-1216,
-1222,-1227,-1232,-1237,-1243,-1249,-1254,-1260,-1266,-1272,-1279,-1285,
-1292,-1299,-1305,-1313,-1320,-1327,-1335,-1343,-1351,-1360,-1369,-1378,
-1387,-1397,-1407,-1417,-1428,-1439,-1451,-1464,-1477,-1490,-1505,-1520,
-1536,-1553,-1572,-1591,-1612,-1636,-1661,-1689,-1720,-1756,-1796,-1845,
-1904,-1980,-2088,-2271 };


/* Timeout generated by Motif monitors exposures and guiding */

void ccd_timeout_handler(client_data_ptr, client_id)
  XtPointer client_data_ptr;
  XtIntervalId client_id;
{
  ccd_status();
  if(exposure_in_progress)
  { 
    read_ccd();
  }
  if(restart)
  {
    start_ccd_exposure();
    get_ds9_region();
  }  

  ccd_timeout_id = XtAppAddTimeOut(context,ccd_timeout_interval,
    ccd_timeout_handler,ccd_timeout_data_ptr);
}  


/* Inquire about the status from the ccd system */

void ccd_status(void)
{
  int  imaging_ccd_status=0, tracking_ccd_status=0;
  QueryCommandStatusParams  in;
  QueryCommandStatusResults  out;
  QueryTemperatureStatusResults  t_out;

  if (camera_type == NO_CAMERA)
  {
    return;
  }   
  in.command = CC_START_EXPOSURE;
  if ((result = SBIGUnivDrvCommand(CC_QUERY_COMMAND_STATUS, &in, &out)) != 0) 
  {
    show_error("status query CC_START_EXPOSURE", result);
  } 
  else 
  {
    imaging_ccd_status = out.status&3;
    tracking_ccd_status = (out.status>>2)&3;
  }
 
  if ((result = SBIGUnivDrvCommand(CC_QUERY_TEMPERATURE_STATUS,NULL,&t_out)) != 0) 
  {
    show_error("status query CC_QUERY_TEMPERATURE_STATUS", result);
  } 
  else 
  {
    ccd_cooler_power=(100./256.)* (double) t_out.power;
    ccd_temperature=0.1*ccd_temp[t_out.ccdThermistor];
    ccd_setpoint_reading =0.1*ccd_temp[t_out.ccdSetpoint];
  }

  show_ccd_status(imaging_ccd_status, tracking_ccd_status);  
}




void ccd_cool_on(void)
{
  SetTemperatureRegulationParams  in;
  double  f;
  int  i,t;

  f = ccd_setpoint_temperature;
  t = 10.0*f + 0.5;

  /* clamp temperature to reasonable range in deg C */

  if (t > 500)
      t = 500;
  else if (t < -500)
      t = -500;
  for (i = 0; i < 4096; ++i)
      if (ccd_temp[i] <= t)
          break;
  in.ccdSetpoint = i;
  in.regulation = 1;

  result = SBIGUnivDrvCommand(CC_SET_TEMPERATURE_REGULATION, &in, NULL);
  if (result != CE_NO_ERROR)
      show_error("CC_SET_TEMPERATURE_REGULATION", result);

}


void ccd_cool_off(void)
{
  SetTemperatureRegulationParams  in;
  double  f;
  int  i,t;

  f = ccd_setpoint_temperature;
  t = 10.0*f + 0.5;

  /* clamp temperature to reasonable range in deg C */

  if (t > 500)
      t = 500;
  else if (t < -500)
      t = -500;
  for (i = 0; i < 4096; ++i)
      if (ccd_temp[i] <= t)
          break;
  in.ccdSetpoint = i;
  in.regulation = 0;

  result = SBIGUnivDrvCommand(CC_SET_TEMPERATURE_REGULATION, &in, NULL);
  if (result != CE_NO_ERROR)
      show_error("CC_SET_TEMPERATURE_REGULATION", result);
}




/* Increment filter */
/* Check bounds based on camera setup */
/* Increment only when the camera is present */

void increment_filter()
{

  if (camera_type == NO_CAMERA) return;
  filter++;
  if (filter > filtermax)
  {
    filter = 1;
  }
}


/* Select a filter and pause for completion */

void set_filter()
{
  int result;
  CFWParams cfw_params;
  CFWResults cfw_results;

  if (camera_type == NO_CAMERA) return;
  if (cfw_model_select == NO_FILTERWHEEL) return;

  cfw_params.cfwCommand = CFWC_GOTO;
  cfw_params.cfwModel = cfw_model_select;
  cfw_params.cfwParam1 = filter;
  
  result = SBIGUnivDrvCommand(CC_CFW, &cfw_params, &cfw_results);
  if (result != CE_NO_ERROR) show_error("CFWC_GOTO", result);
  usleep(5000000);
}


/* Set the AO tip and tilt */

void ao_tip_tilt(int ix, int iy)
{
  AOTipTiltParams  in;

  in.xDeflection=ix;
  in.yDeflection=iy;

  result = SBIGUnivDrvCommand(CC_AO_TIP_TILT, &in, NULL);
  if (result != CE_NO_ERROR)
      show_error("CC_AO_TIP_TILT", result);

}


/* Test the AO by running a square pattern */

void ao_test(int aomag, int aodelay, int aocycles)
{
   int i;
   if (aomag < 0 ) aomag=0;
   if (aomag > 4095) aomag = 4095;
   for (i=0; i<aocycles; ++i )    
   {
     ao_tip_tilt(aomag, aomag);
     usleep(aodelay);
     ao_tip_tilt(aomag, 0);
     usleep(aodelay);
     ao_tip_tilt(0, 0);
     usleep(aodelay);
     ao_tip_tilt(0, aomag);
     usleep(aodelay);
   }
}


/* Activate the relay selected by the value of trackrelay */

void ccd_relay_on(void)
{
  ActivateRelayParams  in;

/* The relays will stay on for 10 seconds (1000 units) unless turned off */

  in.tXPlus=0;
  in.tXMinus=0;
  in.tYPlus=0;
  in.tYMinus=0; 

  switch (trackrelay)
  {
     case 1:
       in.tXPlus=1000;
       break;
     case 2:    
       in.tXMinus=1000;
       break;
     case 4:
       in.tYPlus=1000;
       break;
      case 8:
       in.tYMinus=1000; 
     break;
  }   


  result = SBIGUnivDrvCommand(CC_ACTIVATE_RELAY, &in, NULL);
  if (result != CE_NO_ERROR)
      show_error("CC_ACTIVATE_RELAY", result);

}


/* Deactivate all relays */

void ccd_relay_off(void)
{
  ActivateRelayParams  in;

  in.tXPlus  =  0;
  in.tXMinus =  0;  
  in.tYPlus  =  0;  
  in.tYMinus =  0;
  result = SBIGUnivDrvCommand(CC_ACTIVATE_RELAY, &in, NULL);
  if (result != CE_NO_ERROR)
      show_error("CC_ACTIVATE_RELAY", result);

}


/* Start an exposure controlled by global parameters:                      */
/*    image_mode SINGLE FOCUS MULTI TDI GUIDE                              */
/*    ccd selected by IMAGE (0) or TRACK (1)                               */
/*    track_exposure_time                                                  */
/*    image_exposure_time                                                  */                                  
/*    is_dark                                                              */
/*    exposure_in_progress                                                 */

void start_ccd_exposure(void)
{
  int  result;
  int  exposure_time;
  StartExposureParams  start_exposure_params;
  
  /* Test for exposure in progress */
  
  if( exposure_in_progress==TRUE )
  {
    return;
  }
  
  /* If the exposure is started from the timeout function we need this */
  
  restart=FALSE;

  /* Set up the ccd exposure parameters for the outer loop */
  /* Since image ccd=0 and track ccd=1 this picks one or the other */

  start_exposure_params.ccd = ccd&1;

  /* Use ccd selection to determine exposure time */

  if (ccd==IMAGE)
  {
    exposure_time = (int)(100.0*image_exposure_time + 0.5);  
  }
  else
  {
    exposure_time = (int)(100.0*track_exposure_time);
  }
  start_exposure_params.exposureTime = exposure_time;

  /* User interface does not yet provide for setting the ABG so pick one */

  start_exposure_params.abgState = 0;

  /* Open shutter if this is not a dark frame */

  start_exposure_params.openShutter = (is_dark? 2 : 1);

  /* Send the start request to the camera */

  result = SBIGUnivDrvCommand(CC_START_EXPOSURE,
    &start_exposure_params, NULL);

  /* Test if it was accepted and report any errors */

  if (result != CE_NO_ERROR) 
  {
    show_error("CC_START_EXPOSURE", result);
    return;
  }
  
  /* Start exposure diagnostics */
  
  printf("Exposure started \n");
  
  
  /* The camera is now exposing                                    */
  /* Monitor the camera status for exposure completion             */
  /* Expose and read the track ccd while shutter is open if needed */

  exposure_in_progress=TRUE;
  return;
}


/* Read exposure status                                                    */
/* Action depends on global parameters                                     */
/*    image_mode SINGLE FOCUS MULTI TDI GUIDE                              */
/*    ccd IMAGE (0) or TRACK (1)                                           */
/*    track_exposure_time                                                  */
/*    image_exposure_time                                                  */                                  
/*    is_dark                                                              */
/*    exposure_in_progress                                                 */

void read_ccd(void)
{
  int  result, i, err, status;
  int  inner_exposure_time;
  unsigned short  *image_data;
  StartExposureParams  start_exposure_params;
  StartReadoutParams  start_readout_params;
  EndExposureParams  end_exposure_params;
  MiscellaneousControlParams misc_control_params;
  ReadoutLineParams  readout_line_params;
  QueryCommandStatusParams  query_command_status_params;
  QueryCommandStatusResults  query_command_status_results;
  int readout_x, readout_y, readout_width, readout_height;
  int readout_binning; 
  int ccd_delay; 

  GetCCDInfoResults0  *infop;    

  /* Test that an exposure is actually underway */
  
  if( exposure_in_progress == FALSE )
  {
    return;
  }
     
  if ( (ccd == IMAGE) && (image_mode == TDI) )  
  {

    /* Turn off the restart flag */
    
    restart=FALSE;

    readout_binning = 0;
    infop = &info0;
    readout_x = readout_y = 0;
    readout_width =
      infop->readoutInfo[readout_binning].width;
    readout_height = tdi_height;

    /* Prepare for synchronous line-by-line readout */
    
    end_exposure_params.ccd = ccd&1;                
    SBIGUnivDrvCommand(CC_END_EXPOSURE, &end_exposure_params, NULL);
    
    misc_control_params.fanEnable = TRUE;           /* Fan on */
    misc_control_params.shutterCommand = 1;         /* Shutter open */
    misc_control_params.ledState = 1;               /* LED on */
    SBIGUnivDrvCommand(CC_MISCELLANEOUS_CONTROL, &misc_control_params, NULL);

    start_readout_params.ccd = ccd&1;
    start_readout_params.readoutMode = 0;
    start_readout_params.top = readout_y;
    start_readout_params.left = readout_x;
    start_readout_params.width = readout_width;
    start_readout_params.height = readout_height;
    SBIGUnivDrvCommand(CC_START_READOUT, &start_readout_params, NULL);
    
    sync();
    fflush(stdout);
    ccd_delay = tdi_delay - ccd_line_time;
    if (ccd_delay < 1 ) ccd_delay = 1;
    
    /* TDI diagnostics */ 
    
    printf("TDI started with ccd line delay %d\n", ccd_delay);

    image_data = malloc(readout_width*readout_height*sizeof(unsigned short));
    for (i = 0; i < readout_height; ++i) 
    {
      readout_line_params.ccd = ccd&1;
      readout_line_params.readoutMode = 0;
      readout_line_params.pixelStart = 0;
      readout_line_params.pixelLength = readout_width;
      err = SBIGUnivDrvCommand(CC_READOUT_LINE,
        &readout_line_params,
        image_data + i*readout_width);
      if (err != CE_NO_ERROR) 
      {
        show_error("CC_READOUT_LINE on CCD", err);
        free(image_data);
        return;
      }
      
    /* Pause here for line integration */
      
      usleep(ccd_delay);
    }    

    /* Save the file for ds9 */
    
    write_fits("image.fits", readout_width, readout_height, image_data);

    /* Save an ennumerated file */
    
    sprintf(imagename,"%s%d.fits",basename,imagenumber++);
    write_fits(imagename, readout_width, readout_height, image_data);       
    
    /* Show the tdi image frame on the console */
     
    update_ds9_image();
    free(image_data);

    /* Put the ccd in an idle state */

    misc_control_params.fanEnable = TRUE;           /* Fan on */
    misc_control_params.shutterCommand = 2;         /* Shutter closed */
    misc_control_params.ledState = 0;               /* LED off */
    SBIGUnivDrvCommand(CC_MISCELLANEOUS_CONTROL, &misc_control_params, NULL);

    end_exposure_params.ccd = ccd&1;
    SBIGUnivDrvCommand(CC_END_EXPOSURE, &end_exposure_params, NULL);
    exposure_in_progress = FALSE;

    /* TDIdiagnostics */
  
    printf("TDI completed\n");
  
    return;  
  }

  inner_exposure_time = (int)(100.0*track_exposure_time);
  
  /* Test for an asychronous interrupt from the user */

  if ( interrupt==TRUE ) 
  {

    /* Clear the interrupt flag */

    interrupt = FALSE;

    /* Service the interrupt by terminating the exposure */

    end_exposure_params.ccd = ccd&1;
    SBIGUnivDrvCommand(CC_END_EXPOSURE, &end_exposure_params, NULL);
    
    /* Set the status flag to indicate that the camera is not exposing */
    
    exposure_in_progress=FALSE;
    
    /* Set the restart flag false so that multi exposure mode will stop */
    
    restart=FALSE;
    
    return;
  }

  /* Test whether the CCD exposure is complete  */

  query_command_status_params.command = CC_START_EXPOSURE;
  err = SBIGUnivDrvCommand(CC_QUERY_COMMAND_STATUS,
    &query_command_status_params, &query_command_status_results);
  if ((ccd&1) == 0)
    status = query_command_status_results.status & 3;
  else
    status = (query_command_status_results.status >> 2) & 3;
  if (status == CS_INTEGRATION_COMPLETE) 
  {

    /* CCD exposure complete */

    end_exposure_params.ccd = ccd&1;
    SBIGUnivDrvCommand(CC_END_EXPOSURE, &end_exposure_params, NULL);


    /* Prepare to readout the unbinned image  */

    readout_binning = 0;
    sync();
    fflush(stdout);

    if (ccd == TRACK)
    {
      sprintf(status_buf,"Track readout");
      infop = &info1;
    }
    else
    {
      sprintf(status_buf,"Image readout");
      infop = &info0;
    }
           
    /* Subarea readout permitted in focus mode */

    if ( (image_mode == FOCUS) && (ccd == IMAGE) )
    {  
      readout_x = subarea_x;
      readout_y = subarea_y;
      readout_width =
        infop->readoutInfo[readout_binning].width;
      readout_height =
        infop->readoutInfo[readout_binning].height;

      if ( subarea_width < readout_width )
      {
         readout_width = subarea_width;
      }
      if ( subarea_height < readout_height )
      {
         readout_height = subarea_height;
      }      
    }
    else
    {
      readout_x = readout_y = 0;
      readout_width =
        infop->readoutInfo[readout_binning].width;
      readout_height =
        infop->readoutInfo[readout_binning].height;
    }
    
    /* Read it */

    start_readout_params.ccd = ccd&1;
    start_readout_params.readoutMode = 0;
    start_readout_params.top = readout_y;
    start_readout_params.left = readout_x;
    start_readout_params.width = readout_width;
    start_readout_params.height = readout_height;
    SBIGUnivDrvCommand(CC_START_READOUT, &start_readout_params, NULL);

    image_data = malloc(readout_width*readout_height*sizeof(unsigned short));
    for (i = 0; i < start_readout_params.height; ++i) 
    {
      readout_line_params.ccd = ccd&1;
      readout_line_params.readoutMode = 0;
      readout_line_params.pixelStart = 0;
      readout_line_params.pixelLength = readout_width;
      err = SBIGUnivDrvCommand(CC_READOUT_LINE,
        &readout_line_params,
        image_data + i*readout_width);
      if (err != CE_NO_ERROR) 
      {
        show_error("CC_READOUT_LINE on CCD", err);
        free(image_data);
        return;
      }
    }
    if(ccd == TRACK)
    {
      readout_ccd = TRACK;
      write_fits("track.fits", readout_width, readout_height, image_data);

      /* In image_mode multi also save an ennumerated file */
      
      if( (image_mode == MULTI) )
      {
        sprintf(imagename,"%s%d.fits",basename,imagenumber++);
        write_fits(imagename, readout_width, readout_height, image_data);       
      }
            
      /* Show the latest track frame on the console */
      
      update_ds9_track();
    }
    else
    {
      readout_ccd = IMAGE;
      write_fits("image.fits", readout_width, readout_height, image_data);

      /* In multi-image mode also save an ennumerated file */
      
      if( image_mode == MULTI )
      {
        sprintf(imagename,"%s%d.fits",basename,imagenumber++);
        write_fits(imagename, readout_width, readout_height, image_data);       
      }
      
      /* Show the latest image frame on the console */
       
      update_ds9_image();
    }

    if( image_mode != SINGLE )
    {
      restart=TRUE;
    }
    else
    {
      restart=FALSE;
    }
    free(image_data);
    exposure_in_progress = FALSE;
    return;
  }

  /* Expose and read the track ccd in guide mode */

  if ( (ccd == IMAGE) && (image_mode == GUIDE) )  
  {
    unsigned short  *tracking_data;
    int  width, height;

    width = info1.readoutInfo[0].width;
    height = info1.readoutInfo[0].height;
    tracking_data = malloc(width*height*sizeof(unsigned short));
    start_exposure_params.ccd=TRACK;
    start_exposure_params.exposureTime = inner_exposure_time;
    start_exposure_params.abgState=ABG_CLK_MED7;
    start_exposure_params.openShutter = 0;  /* leave shutter as is */
    result = SBIGUnivDrvCommand(CC_START_EXPOSURE,
                           &start_exposure_params, NULL);
    if (result != CE_NO_ERROR) 
    {
      show_error("CC_START_EXPOSURE", result);
       
      /* End outer exposure ... */

     end_exposure_params.ccd = 0;
     SBIGUnivDrvCommand(CC_END_EXPOSURE, &end_exposure_params, NULL);
     free(tracking_data);
     return;
    }

    /* Wait here for selfguide exposure to complete */

    while (TRUE) 
    {
      query_command_status_params.command = CC_START_EXPOSURE;
      SBIGUnivDrvCommand(CC_QUERY_COMMAND_STATUS,
        &query_command_status_params, &query_command_status_results);
      status = (query_command_status_results.status >> 2) & 3;
      if (status == CS_INTEGRATION_COMPLETE)
        break;
    }
    
    /* End self-guiding track ccd exposure */
    
    end_exposure_params.ccd = TRACK;
    SBIGUnivDrvCommand(CC_END_EXPOSURE, &end_exposure_params, NULL);

    /* Read out only the tracking ccd */
    
    start_readout_params.ccd = TRACK;
    start_readout_params.readoutMode = 0;
    start_readout_params.top = 0;
    start_readout_params.left = 0;
    start_readout_params.width = width;
    start_readout_params.height = height;
    SBIGUnivDrvCommand(CC_START_READOUT, &start_readout_params, NULL);
    for (i = 0; i < start_readout_params.height; ++i) 
    {
      readout_line_params.ccd = TRACK;
      readout_line_params.readoutMode = 0;
      readout_line_params.pixelStart = 0;
      readout_line_params.pixelLength = width;
      err = SBIGUnivDrvCommand(CC_READOUT_LINE,
                    &readout_line_params, tracking_data + i*width);
      if (err != CE_NO_ERROR) 
      {
        show_error("CC_READOUT_LINE on tracking CCD", err);
        free(tracking_data);
        return;
      }
    }
    
    /* Self-guiding feedback control goes here */
    
    /* Write track image to the same file each time */
    
    readout_ccd = TRACK;
    write_fits("track.fits", width, height, tracking_data);
    free(tracking_data);
    
    /* Show the latest tracking image */
 
    update_ds9_track();
  } 
}


/* Start ds9 as a new process */

void start_ds9(void)
{
  strcpy(buf, "ds9 ds9.fits &");
  system(buf);
}

       
/* Use xpa to tell ds9 to read image.fits */

void update_ds9_image(void)
{  
  strcpy(buf, "xpaset ds9 fits ds9.fits < image.fits");
  system(buf);
}


/* Use xpa to tell ds9 to read track.fits */

void update_ds9_track(void)
{  
  strcpy(buf, "xpaset ds9 fits ds9.fits < track.fits");
  system(buf);
}


/* Use xpa to read last region from ds9 */

void get_ds9_region(void)
{  
  char regionstr[121];
  char *regionptr=regionstr;
  FILE *fp_region;
  int j,k;
  
  strcpy(buf, "xpaget ds9 regions > region.txt");
  system(buf);
  fp_region=fopen("region.txt", "r");
  while ( regionstr == fgets(regionstr,120,fp_region) )
  {    
    k=strlen(regionstr);
    if (k>1)
    {
      
      /* Prepare the string by writing a newline character at the end */
      
      regionstr[k]='\0';
   
      /* Set the pointer to the start of the string */

      regionptr=regionstr;
            
      /* Point to the last region */
      /* Format is circle(x_center,y_center,radius) */
      
      regionptr=strrchr(regionptr, '(' );
      if(regionptr!=NULL)
      {
        regionptr++;
        j=sscanf(regionptr,"%lf,%lf,%lf,%lf",
          &region_x,&region_y,&region_dx,&region_dy);
        
        /* Print region diagnostics */
        
        /* printf("%lf %lf %lf %lf\n",region_x,region_y,region_dx,region_dy); */
        
      }           

    } 
 
  }
  fclose(fp_region);
  
}



/* Print a list of command line arguments */

static void usage(FILE *stream)
{
    fprintf(stream, "\n");
    fprintf(stream, "Usage of xmccd command line arguments: \n");
    fprintf(stream, "  -usb  Camera connected to a usb port\n");    
    fprintf(stream, "  -lpt1 Camera connected to first parallel port\n");
    fprintf(stream, "  -lpt2 Camera connected to second parallel port\n");
    fprintf(stream, "  -lpt3 Camera connected to third parallel port\n");
    fprintf(stream, "\n");
    exit(0);
}

/* Compare arguments on the command line with an expected response */

int argmatch(char *s1, char *s2, int atleast)
{
    int l1 = strlen(s1);
    int l2 = strlen(s2);

    if (l1 < atleast) {
        return 0;
    }
    if (l1 > l2) {
        return 0;
    }
    return (strncmp(s1, s2, l1) == 0);
}
           

/* Print out cfitsio error messages and exit program */


void show_cfitsio_error(int status)
{
    if (status)
    {
       fits_report_error(stderr, status); /* print error report */

       exit( status );    /* terminate the program, returning error status */
    }
    return;
}
